/*++++++++++++++++++++++
 refdbdref.c: refdb application server, reference handling functions 
 markus@mhoenicka.de 2-10-00 
 $Id: refdbdref.c,v 1.65.2.23 2005/11/03 20:40:19 mhoenicka Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   ++++++++++++++++++++++++*/

/* ToDo: duplicate check. Could work like this:
   new command checkref sends datasets like addref/updateref
   but adds the refs to a temporary database
   refdbd then checks
   - fuzzy title search
   - JO,PY,VL,IS,SP
   - fuzzy author search
   - fuzzy journal name search

   Output is a screen or html report of possible duplicate references,
   possible duplicate author names, and possible journal name duplicates
*/

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <syslog.h> /* priority level definitions */
#include <time.h>
#include <expat.h> /* the xml parser header file */
#include <iconv.h>
#include <errno.h>
#include <dbi/dbi.h>

#include "strfncs.h"
#include "refdb.h"
#include "connect.h"
#include "backend.h"
#include "linklist.h"
#include "refdbd.h" /* depends on backend.h */
#include "tokenize.h"
#include "readris.h"
#include "writeris.h" /* depends on backend.h */
#include "risdb.h"
#include "dbfncs.h"
#include "xmlhandler.h"
#include "risxhandler.h"
#include "authorinfo.h"
#include "risdata.h"

/* #define _GNU_SOURCE */ /* need this for long long support on some platforms */

/* some globals */
extern char server_ip[]; /* the IP address of the database server */
extern char refdblib[]; /* location of shareable data */
extern int n_log_level; /* numeric version of log_level */
extern int nongeek_offset; /* :-) */
extern char cs_term[];

/* forward declaration of local functions */
static unsigned long long read_ris_data(struct CLIENT_REQUEST* ptr_clrequest, struct ADDRESULT* ptr_addresult, dbi_conn conn, dbi_conn conn_refdb, int replace_ref, int n_keep_id, const char* the_user, Lilid* ptr_sentinel, iconv_t conv_descriptor, int iconv_init_status);
static int real_run_keyword_scan(struct CLIENT_REQUEST* ptr_clrequest, Lilid* ptr_sentinel, int mode, int* ptr_insert);
static int is_journal(dbi_conn conn, const char* field, const char* quoted_name);


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  update_meta(): updates the meta information in a reference database

  int update_meta returns 1 if failed, 0 if successful

  dbi_conn conn connection to a database

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int update_meta(dbi_conn conn, struct CLIENT_REQUEST* ptr_clrequest) {
  dbi_result dbires;
  char date_buffer[24];
  char sql_command[128];
  time_t the_time;

  time(&the_time);
  strftime(date_buffer, 24, "%Y-%m-%d %H:%M:%S", gmtime(&the_time));

  if (!strcmp(my_dbi_conn_get_cap(conn, "multiple_db"), "f")) {
    sprintf(sql_command, "UPDATE t_meta SET meta_modify_date='%s'", date_buffer);
  }
  else {
    sprintf(sql_command, "UPDATE %s.t_meta SET meta_modify_date='%s'", ptr_clrequest->current_db, date_buffer);
  }

  LOG_PRINT(LOG_DEBUG, sql_command);

  dbires = dbi_conn_query(conn, sql_command);
  if (!dbires) {
    /* treat as non-critical */
    LOG_PRINT(LOG_WARNING, "updating t_meta failed");
    return 1;
  }

  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  whichdb(): implements the client command whichdb

  int whichdb returns 1 if failed, 0 if successful

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int whichdb(struct CLIENT_REQUEST* ptr_clrequest) {
  dbi_conn conn;
  dbi_driver driver;
  dbi_result dbires;
  dbi_error_flag errflag;
  char create_date[64];
  char modify_date[64];
  char* return_msg;
  const char* db_encoding;
  unsigned long long numrefs;
  unsigned long long maxid;
  unsigned long long numnotes;
  unsigned long long maxnid;
  unsigned short fieldtype;
  int numbyte;
  int error;
  time_t the_time;

  /* get some memory for a return message */
  return_msg = malloc((size_t)512);
  if (return_msg == NULL) {
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 1;
  }

  /* connect to database server*/
  if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) == NULL) {
    send_status(ptr_clrequest->fd, 204, TERM_NO);
    LOG_PRINT(LOG_WARNING, get_status_msg(204));
    free(return_msg);
    return 1;
  }

  driver = dbi_conn_get_driver(conn);

  db_encoding = dbi_conn_get_encoding(conn);

  dbires = dbi_conn_query(conn, "SELECT COUNT(*), MAX(refdb_id) FROM t_refdb WHERE refdb_type!='DUMMY'");
  if (!dbires
      || !dbi_result_next_row(dbires)) {
    send_status(ptr_clrequest->fd, 207, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(207));
    if (dbires) {
      dbi_result_free(dbires);
    }
    dbi_conn_close(conn);
    free(return_msg);
    return 1;
  }
    
  numrefs = my_dbi_result_get_idval_idx(dbires, 1); /* 1-base index */
  maxid = my_dbi_result_get_idval_idx(dbires, 2);
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "SELECT COUNT(*), MAX(note_id) FROM t_note");
  if (!dbires
      || !dbi_result_next_row(dbires)) {
    send_status(ptr_clrequest->fd, 207, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(207));
    if (dbires) {
      dbi_result_free(dbires);
    }
    dbi_conn_close(conn);
    free(return_msg);
    return 1;
  }
    
  numnotes = my_dbi_result_get_idval_idx(dbires, 1); /* 1-base index */
  maxnid = my_dbi_result_get_idval_idx(dbires, 2);
  dbi_result_free(dbires);

  dbires = dbi_conn_query(conn, "SELECT meta_app,meta_type,meta_version,meta_dbversion,meta_create_date,meta_modify_date FROM t_meta");
  if (!dbires
      || !dbi_result_next_row(dbires)) {
    send_status(ptr_clrequest->fd, 207, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(207));
    if (dbires) {
      dbi_result_free(dbires);
    }
    dbi_conn_close(conn);
    free(return_msg);
    return 1;
  }

  the_time = dbi_result_get_datetime(dbires, "meta_create_date");
  if (dbi_conn_error_flag(conn) == 0) {
    if (the_time) {
      strftime(create_date, 64, "%Y-%m-%d %H:%M:%S", gmtime(&the_time));
    }
    else {
      strcpy(create_date, "nd");
    }
  }
  else {
    strcpy(create_date, "nd");
  }

  the_time = dbi_result_get_datetime(dbires, "meta_modify_date");
  if (dbi_conn_error_flag(conn) == 0) {
    if (the_time) {
      strftime(modify_date, 64, "%Y-%m-%d %H:%M:%S", gmtime(&the_time));
    }
    else {
      strcpy(modify_date, "nd");
    }
  }
  else {
    strcpy(modify_date, "nd");
  }

  /* The returned integers are quads (8byte) */
  sprintf(return_msg, "Current database: %s\nNumber of references: "ULLSPEC"\nHighest reference ID: "ULLSPEC"\nNumber of notes: "ULLSPEC"\nHighest note ID: "ULLSPEC"\nEncoding: %s\nDatabase type: %s\nDatabase version: %d\nDatabase server: %s\nCreated: %s UTC\nUsing %s version: %s\nLast modified: %s UTC\n", ptr_clrequest->current_db, (unsigned long long)numrefs, (unsigned long long)maxid, (unsigned long long)numnotes, (unsigned long long)maxnid, db_encoding, dbi_result_get_string(dbires, "meta_type"),dbi_result_get_short(dbires, "meta_dbversion"),  dbi_driver_get_name(driver), create_date, dbi_result_get_string(dbires, "meta_app"), dbi_result_get_string(dbires, "meta_version"), modify_date);

  dbi_result_free(dbires);

  send_status(ptr_clrequest->fd, 0, TERM_NO);
  numbyte = tiwrite(ptr_clrequest->fd, return_msg, TERM_YES);
  if (numbyte == -1) {
    LOG_PRINT(LOG_WARNING, "Cannot write to client");
  }

  free(return_msg);

  dbi_conn_close(conn);

  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  addref(): implements the client command addref

  int addref returns 1 if failed, 0 if successful

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  char *set_owner the username of the owner of new/updated
                      references if different from *username

  struct ADDRESULT* addresult this structure will be filled in with the number
                      of (un-)successfully added/updated references

  int replace_ref if 1, ref will be updated according to ID field; if 0
         ref will be added without looking at an ID field; if 2, only the
         personal settings (N1, AV, RP) will be updated according to the
         ID field

  int n_keep_id if 1, any existing reference ID will be saved in U5

  const char* reftype ptr to string with reference type (ris|risx)

  Lilid* ptr_sentinel ptr to linked list that will receive all
                      added ID values

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int addref(struct CLIENT_REQUEST* ptr_clrequest, char* set_owner, struct ADDRESULT* ptr_addresult, int replace_ref, int n_keep_id, const char* reftype, Lilid* ptr_sentinel) {
  dbi_conn conn;
  dbi_conn conn_refdb = NULL;
  dbi_result dbires;
  XML_Parser p;
  int numbyte; /* number of bytes written */
  unsigned long long ullresult; /* result value of processing a RIS set */
  int retval = 0; /* be optimistic */
  int iconv_init_status = 0; /* indicates iconv initialization problem */
  char *the_user; /* name (either set_owner or username) */
  char *return_msg; /* string to hold a return message for the client */
  const char *db_encoding; /* encoding of database */
  const char *drivername; /* name of the libdbi driver */
  char old_db[DBNAME_LENGTH+1];
  struct addrisx_data ardata;
  struct lilimem sentinel;

  sentinel.ptr_mem = NULL;
  sentinel.ptr_next = NULL;
  sentinel.varname[0] = '\0';

  ardata.msgpool = NULL;
  ardata.ptr_ainfo = NULL;

  /* if set_owner is set, use it, otherwise use username as owner */
  if (set_owner && *set_owner) {
    the_user = set_owner;
  }
  else {
    the_user = ptr_clrequest->username;
  }

  /* get some memory for a return message */
  return_msg = malloc((size_t)256);
  if (return_msg == NULL || insert_lilimem(&sentinel, (void**)&return_msg, NULL)) {
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 1;
  }

  /* connect to database server*/
  if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) == NULL) {
    send_status(ptr_clrequest->fd, 204, TERM_NO);
    LOG_PRINT(LOG_WARNING, get_status_msg(204));
    delete_all_lilimem(&sentinel);
    return 1;
  }

  /* get the database encoding */
  db_encoding = dbi_conn_get_encoding(conn);

  /* if we need to convert, create a conversion descriptor for iconv() */
  if (db_encoding && strcmp(db_encoding, !strcmp(reftype, "risx") ? "UTF-8" : ptr_clrequest->db_encoding)) {
    ardata.conv_descriptor = iconv_open(!strcmp(db_encoding, "US-ASCII") ? "ASCII" : db_encoding, !strcmp(reftype, "risx") ? "UTF-8" : ptr_clrequest->db_encoding);
    if (ardata.conv_descriptor == (iconv_t)(-1)) {
      ardata.conv_descriptor = NULL;
      LOG_PRINT(LOG_WARNING, "cannot set conversion descriptor (input/database):");
      LOG_PRINT(LOG_DEBUG, ptr_clrequest->db_encoding);
      LOG_PRINT(LOG_DEBUG, db_encoding);
      iconv_init_status = 701;
    }
    else {
      LOG_PRINT(LOG_DEBUG, "input encoding is:");
      LOG_PRINT(LOG_DEBUG, ptr_clrequest->db_encoding);
    }
  }
  else {
    ardata.conv_descriptor = NULL;
    LOG_PRINT(LOG_DEBUG, "no character encoding conversion required");
  }

  LOG_PRINT(LOG_DEBUG, "database encoding is:");
  LOG_PRINT(LOG_DEBUG, db_encoding);

  if (!strcmp(my_dbi_conn_get_cap(conn, "multiple_db"), "f")) {
    /* pgsql, sqlite, sqlite3 need another connection to retrieve values from
       the refdb database */

    conn_refdb = connect_to_db(ptr_clrequest, MAIN_DB, 0);

    if (conn_refdb == NULL) {
      send_status(ptr_clrequest->fd, 202, TERM_NO);
      LOG_PRINT(LOG_WARNING, get_status_msg(202));
      dbi_conn_close(conn);
      delete_all_lilimem(&sentinel);
      if (ardata.conv_descriptor) {
	iconv_close(ardata.conv_descriptor);
      }
      return 1;
    }
  }
  /* else: do nothing for other drivers */

  if (!strcmp(reftype, "risx")) {
    while (1) { /* leave with break */
      /* create the parser instance */
      p = XML_ParserCreate(NULL);
      if (!p) {
	send_status(ptr_clrequest->fd, 801, TERM_NO);
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	retval = 1;
	goto Finish;
      }

      /* initialize "global" handler data */
      ardata.msgpool_len = 512;
      ardata.msgpool = malloc(ardata.msgpool_len);
      if (!ardata.msgpool || insert_lilimem(&sentinel, (void**)&(ardata.msgpool), NULL)) {
	send_status(ptr_clrequest->fd, 801, TERM_NO);
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	retval = 1;
	goto Finish;
      }

      ardata.ptr_ainfo = new_authorinfo();
      if (!ardata.ptr_ainfo) {
	send_status(ptr_clrequest->fd, 801, TERM_NO);
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	retval = 1;
	goto Finish;
      }

      *(ardata.msgpool) = '\0';
      *(ardata.real_citekey) = '\0';
      *(ardata.type) = '\0';
      strncpy(ardata.username, the_user, 16);
      ardata.username[15] = '\0';
      ardata.ptr_clrequest = ptr_clrequest;
      ardata.ptr_id_sentinel = ptr_sentinel;
      ardata.ptr_first = NULL;
      ardata.ptr_risdata = NULL;
      ardata.nmem_error = 0;
      ardata.ndb_error = 0;
      ardata.n_skip = 0;
      ardata.depth = 0;
      ardata.depth_adjust = 0;
      ardata.set_count = 0;
      ardata.added_count = 0;
      ardata.skipped_count = 0;
      ardata.updated_count = 0;
      ardata.failed_count = 0;
      ardata.authorpos = 0;
      ardata.replace_ref = replace_ref;
      ardata.create_new = 1;
      ardata.n_user_id = 0;
      ardata.conn = conn;
      ardata.conn_refdb = conn_refdb;
      ardata.driver = dbi_conn_get_driver(conn);
      ardata.drivername = dbi_driver_get_name(ardata.driver);


      /* register our handlers. these handlers will be called whenever expat finds a start- or endtag or character data */
      XML_SetElementHandler(p, risx_start_handler, risx_end_handler);
      XML_SetCharacterDataHandler(p, risx_char_handler);
      
      /* make pointer to "global data" available for handlers */
      XML_SetUserData(p, (void*)&ardata);

      send_status(ptr_clrequest->fd, iconv_init_status, TERM_NO);

      ullresult = read_xml(ptr_clrequest->fd, p, ptr_addresult);
      XML_ParserFree(p);

      /* do something intelligent in the case of a parse or mem error */
      if (!ullresult || ardata.ndb_error || ardata.nmem_error) {
	if (strcmp(my_dbi_conn_get_cap(ardata.conn, "transaction"), "t")) {
	  /* we have to delete the junk reference manually */
	  delete_ref_by_id(ardata.n_refdb_id, conn, ptr_clrequest, ptr_addresult);
	}
	else {
	  my_dbi_conn_rollback(ardata.conn);
	}
	my_dbi_conn_unlock(ardata.conn);
      }

      /* send messages to client */
/* 	send_status(ptr_clrequest->fd, 400, TERM_NO); */
      numbyte = tiwrite(ptr_clrequest->fd, ardata.msgpool, TERM_YES);
      if (numbyte == -1) {
	LOG_PRINT(LOG_INFO, "timeout while writing");
	retval = 1;
	goto Finish;
      }

      ptr_addresult->success += ardata.added_count;
      ptr_addresult->failure += ardata.failed_count;
      ptr_addresult->updated += ardata.updated_count;
      ptr_addresult->skipped += ardata.skipped_count;
      if (!ullresult) { /* error */
	retval = 1;
	goto Finish;
      }
      else if (ullresult == -1) { /* we're done */
	break;
      }
    } /* end while */
  }
  else { /* ris */
    ullresult = read_ris_data(ptr_clrequest, ptr_addresult, conn, conn_refdb, replace_ref, n_keep_id, the_user, ptr_sentinel, ardata.conv_descriptor, iconv_init_status);

    /* terminate whatever was sent so far */
    numbyte = tiwrite(ptr_clrequest->fd, "", TERM_YES);
    if (numbyte == -1) {
      LOG_PRINT(LOG_INFO, "timeout while writing");
      retval = 1;
    }
  }

 Finish:
  if (ptr_addresult->success || ptr_addresult->updated) {
    /* database was changed, update meta info */
    update_meta(conn, ptr_clrequest);
  }
  
  if (conn_refdb) {
    dbi_conn_close(conn_refdb);
  }
  dbi_conn_close(conn);
  delete_all_lilimem(&sentinel);
  free_authorinfo(ardata.ptr_ainfo);

  if (ardata.conv_descriptor) {
    iconv_close(ardata.conv_descriptor);
  }

  return retval;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  read_ris_data(): reads RIS data from the client

  static unsigned long long read_ris_data returns 0 if failed, >0  if successful

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  struct ADDRESULT* addresult this structure will be filled in with the number
                      of (un-)successfully added/updated references

  dbi_conn conn connection to reference database

  dbi_conn conn_refdb connection to shared database
  
  int replace_ref if 1, ref will be updated according to ID field; if 0
         ref will be added without looking at an ID field; if 2, only the
         personal settings (N1, AV, RP) will be updated according to the
         ID field

  int n_keep_id if 1, any existing reference ID will be saved in U5

  char *the_user the username of the owner of new/updated
                      references

  Lilid* ptr_sentinel ptr to linked list that will receive all
                      added ID values

  iconv_t conv_descriptor conversion descriptor for iconv()

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static unsigned long long read_ris_data(struct CLIENT_REQUEST* ptr_clrequest, struct ADDRESULT* ptr_addresult, dbi_conn conn, dbi_conn conn_refdb, int replace_ref, int n_keep_id, const char* the_user, Lilid* ptr_sentinel, iconv_t conv_descriptor, int iconv_init_status) {
  int numbyte;
  int success = 1; /* be optimistic */
  int n_requested_bufsize; /* buffer size a client requests to transfer data */
  int result; /* result value of processing a RIS set */
  int cs_status;
  unsigned long long set_count = 0; /* counter for datasets */
  char inbuffer[COMMAND_INBUF_LEN]; /* buffer for client commands */
  char return_msg[256];
  char *ris_set; /* pointer to a string containing a RIS dataset */
  char *new_ris_set; /* pointer used during reallocation of *ris_set */
  char *msg_pool; /* collects messages during RIS parsing */
  size_t msg_pool_len;

  ris_set = malloc((size_t)1);
  if (ris_set == NULL) {
    /* out of memory */
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 0;
  }

  msg_pool_len = 256;
  msg_pool = malloc(msg_pool_len);
  if (msg_pool == NULL) {
    /* out of memory */
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    free(ris_set);
    return 0;
  }

  send_status(ptr_clrequest->fd, iconv_init_status, TERM_NO);

  /* loop until the client requests a buffer size of zero indicating
     that he's done sending datasets */
  do {
    /* ------------------------------------------------------------ */
    /* PHASE 1 */
    /* see how much memory we'll need */
/*     printf("phase 1\n"); */
    
    cs_status = read_status(ptr_clrequest->fd);
    if (cs_status == 402) {
      /* client is done sending data */
      send_status(ptr_clrequest->fd, 403, TERM_NO);
      LOG_PRINT(LOG_INFO, get_status_msg(403));
      free(ris_set);
      free(msg_pool);
/*       printf("leaving after 402\n"); */
      break;
    }

    /* see how much memory we'll need */
    numbyte = tread(ptr_clrequest->fd, inbuffer, 10);
    if (numbyte == -1) {
      LOG_PRINT(LOG_INFO, get_status_msg(109));
      free(ris_set);
      free(msg_pool);
      return 0;
    }
    
    /* the size of the chunk the client wants to send */
    n_requested_bufsize = atoi(inbuffer);
/*     printf("inbuffer: %s\nbufsize: %d\n", inbuffer, n_requested_bufsize); */

    /* ------------------------------------------------------------ */
    /* PHASE 2 */
    /* send back reply */
/*     printf("phase 2\n"); */

    /* try to get a buffer large enough to hold the reference
       if n_requested_bufsize == 0, then this command equals free() */
    new_ris_set = (char*)realloc(ris_set, (size_t)n_requested_bufsize);
    if (new_ris_set == NULL && n_requested_bufsize) {
      send_status(ptr_clrequest->fd, 801, TERM_NO);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      free(ris_set);
      free(msg_pool);
      return 0;
    }
    else {
      ris_set = new_ris_set;
    }

    /* read the data proper unless size was zero (= end of transmission) */
    if (n_requested_bufsize) {
      size_t inlength;
      size_t outlength;
      char* my_ris_set = NULL; /* this ptr will be modified by iconv() */
      char* my_ris_set_start = NULL; /* records initial state of my_elvalue */
      const char* my_instring = NULL; /* this ptr will be modified by iconv() */

      /* acknowledge that we have a buffer ready, let them data come... */
      send_status(ptr_clrequest->fd, 0, TERM_NO);

      /* ------------------------------------------------------------ */
      /* PHASE 3 */
      /* read dataset */
/*       printf("phase 3\n"); */
      numbyte = iread(ptr_clrequest->fd, ris_set, n_requested_bufsize);
      /*        printf("requested: %d\nread: %d\n", n_requested_bufsize, numbyte); */
      if (numbyte == -1) {
	LOG_PRINT(LOG_INFO, get_status_msg(109));
	free(ris_set);
	free(msg_pool);
	return 0;
      }

      /* run a character encoding conversion if required */
      if (conv_descriptor && numbyte) {
	inlength = numbyte - TERM_LEN;
	/* with the encodings supported by our database engines, the converted
	   string can't be longer than six times the input string */
	outlength = 6*inlength;

	if ((my_ris_set = malloc(outlength)) == NULL) {
	  ptr_addresult->failure++;
	  send_status(ptr_clrequest->fd, 801, TERM_NO);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  free(msg_pool);
	  return 0;
	}

	/* keep start of the converted string */
	my_ris_set_start = my_ris_set;

	/* variable will be modified by iconv, so don't use original */
	my_instring = (const char*)ris_set;

	/* now actually do the conversion */
	if (iconv(conv_descriptor, &my_instring, &inlength, &my_ris_set, &outlength) == (size_t)(-1)) {
	  if (errno == EILSEQ) {
	    sprintf(return_msg, "iconv: invalid input character sequence\n");
	    LOG_PRINT(LOG_WARNING, "iconv: invalid input character sequence");
	  }
	  else if (errno == E2BIG) {
	    sprintf(return_msg, "iconv: output buffer too small\n");
	    LOG_PRINT(LOG_WARNING, "iconv: output buffer too small");
	  }
	  else if (errno == EINVAL) {
	    sprintf(return_msg, "iconv: incomplete input character\n");
	    LOG_PRINT(LOG_WARNING, "iconv: incomplete input character");
	  }
	  
	  send_status(ptr_clrequest->fd, 702, TERM_NO);
	  LOG_PRINT(LOG_CRIT, get_status_msg(702));
	  free(ris_set);
	  free(msg_pool);
	  return 0;
	}
	/* else: conversion went ok. We free the original string and replace
	   it with the converted copy */
	if (ris_set) {
	  free(ris_set);
	}
	ris_set = my_ris_set_start;
      }
      /* else: no conversion required */

      /* ------------------------------------------------------------ */
      /* PHASE 4 */
      /* send back message */
/*       printf("phase 4\n"); */

      /* split reference into lines and squeeze contents into database */
/*       printf("entering process_ris_set()\n"); */
      
      *msg_pool = '\0'; /* reset */
      result = process_ris_set(ris_set, conn, conn_refdb, replace_ref, (char*)the_user, ptr_clrequest, n_keep_id, ptr_sentinel, set_count, &msg_pool, &msg_pool_len);
/*       printf("leaving process_ris_set()\n"); */
      if (result == 1) {
	ptr_addresult->success++;
/* 	sprintf(return_msg, "Adding input set %d successful\n", set_count + nongeek_offset); */
	send_status(ptr_clrequest->fd, 408, TERM_NO);
	if (*msg_pool) {
	  numbyte = tiwrite(ptr_clrequest->fd, msg_pool, TERM_NO);
	  if (numbyte == -1) {
	    LOG_PRINT(LOG_INFO, get_status_msg(110));
	    free(ris_set);
	    free(msg_pool);
	    return 0;
	  }
	}
	sprintf(return_msg, "408:"ULLSPEC"\n", set_count + nongeek_offset);
	LOG_PRINT(LOG_INFO, "dataset added successfully");
	numbyte = tiwrite(ptr_clrequest->fd, return_msg, TERM_YES);
	if (numbyte == -1) {
	  LOG_PRINT(LOG_INFO, get_status_msg(110));
	  free(ris_set);
	  free(msg_pool);
	  return 0;
	}
      }
      else if (result == 2) {
	ptr_addresult->updated++;
/* 	sprintf(return_msg, "Updating input set %d successful\n", set_count + nongeek_offset); */
	send_status(ptr_clrequest->fd, 413, TERM_NO);
	if (*msg_pool) {
	  numbyte = tiwrite(ptr_clrequest->fd, msg_pool, TERM_NO);
	  if (numbyte == -1) {
	    LOG_PRINT(LOG_INFO, get_status_msg(110));
	    free(ris_set);
	    free(msg_pool);
	    return 0;
	  }
	}
	sprintf(return_msg, "413:"ULLSPEC"\n", set_count + nongeek_offset);
	LOG_PRINT(LOG_INFO, "dataset updated successfully");
	numbyte = tiwrite(ptr_clrequest->fd, return_msg, TERM_YES);
	if (numbyte == -1) {
	  LOG_PRINT(LOG_INFO, get_status_msg(110));
	  free(ris_set);
	  free(msg_pool);
	  return 0;
	}
      }
      else if  (result == 3) {
	ptr_addresult->skipped++;
	send_status(ptr_clrequest->fd, 407, TERM_NO);
	if (*msg_pool) {
	  numbyte = tiwrite(ptr_clrequest->fd, msg_pool, TERM_NO);
	  if (numbyte == -1) {
	    LOG_PRINT(LOG_INFO, get_status_msg(110));
	    free(ris_set);
	    free(msg_pool);
	    return 0;
	  }
	}
	sprintf(return_msg, "407:"ULLSPEC"\n", set_count + nongeek_offset);
	LOG_PRINT(LOG_INFO, "skipped processing dataset");
	numbyte = tiwrite(ptr_clrequest->fd, return_msg, TERM_YES);
	if (numbyte == -1) {
	  LOG_PRINT(LOG_INFO, get_status_msg(110));
	  free(ris_set);
	  free(msg_pool);
	  return 0;
	}
      }
      else {
	ptr_addresult->failure++;
	send_status(ptr_clrequest->fd, 414, TERM_NO);
	if (*msg_pool) {
	  numbyte = tiwrite(ptr_clrequest->fd, msg_pool, TERM_NO);
	  if (numbyte == -1) {
	    LOG_PRINT(LOG_INFO, get_status_msg(110));
	    free(ris_set);
	    free(msg_pool);
	    return 0;
	  }
	}
	sprintf(return_msg, "414:"ULLSPEC"\n", set_count + nongeek_offset);
	LOG_PRINT(LOG_WARNING, "failed processing dataset");
	numbyte = tiwrite(ptr_clrequest->fd, return_msg, TERM_YES);
	if (numbyte == -1) {
	  LOG_PRINT(LOG_INFO, get_status_msg(110));
	  free(ris_set);
	  free(msg_pool);
	  return 0;
	}
      }
      set_count++;
    }
  } while (n_requested_bufsize);

  /* ris_set should be deallocated by now due to a realloc(0) call */
/*   printf("leaving at end\n"); */
  return set_count;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  run_keyword_scan(): scans datasets for keywords after adding
                      (dispatcher for real_run_keyword_scan())

  int run_keyword_scan returns 1 if failed, 0 if successful

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  Lilid* ptr_sentinel ptr to linked list with ID values to check
                      or NULL if all references in the database
		      should be scanned

  int mode 0 = reference entries 1 = note entries 2 = both
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int run_keyword_scan(struct CLIENT_REQUEST* ptr_clrequest, Lilid* ptr_sentinel, int mode) {
  int retval;
  int n_insert = 0; /* number of inserts that actually ran */
  dbi_conn conn;

  if (!mode) { /* check only references */
    retval = real_run_keyword_scan(ptr_clrequest, ptr_sentinel, 0, &n_insert);
  }
  else if (mode == 1) { /* check only notes */
    retval = real_run_keyword_scan(ptr_clrequest, ptr_sentinel, 1, &n_insert);
  }
  else { /* check both references and notes */
    retval = real_run_keyword_scan(ptr_clrequest, ptr_sentinel, 0, &n_insert);
    
    if (!retval) {
      retval = real_run_keyword_scan(ptr_clrequest, ptr_sentinel, 1, &n_insert);
    }
  }
  
  if (n_insert) {
    /* connect to database server*/
    if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) == NULL) {
      LOG_PRINT(LOG_ERR, get_status_msg(205));
      return 1;
    }
    update_meta(conn, ptr_clrequest);

    dbi_conn_close(conn);
  }

  return retval;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  real_run_keyword_scan(): scans datasets for keywords after adding

  int real_run_keyword_scan returns 1 if failed, 0 if successful

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  Lilid* ptr_sentinel ptr to linked list with ID values to check
                      or NULL if all references in the database
		      should be scanned

  int mode 0 = reference entries 1 = note entries

  int* ptr_insert ptr to insert counter

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int real_run_keyword_scan(struct CLIENT_REQUEST* ptr_clrequest, Lilid* ptr_sentinel, int mode, int* ptr_insert) {
  int i;
  unsigned long long keyword_id;
  unsigned long long xref_id;
  size_t sql_cmd_len = 1024;
  char *sql_command;
  char *new_sql_command;
  char id_buffer[21] = "";
  const char *item;
  const char *item_kw;
  const char scanfields[2][5][19] = {
    {"refdb_abstract", "refdb_title", "refdb_booktitle", "refdb_title_series", ""},
    {"note_title", "note_content", ""}};

  dbi_conn conn;
  dbi_result dbires;
  dbi_result dbires1;
  dbi_result dbires2;
  dbi_result dbires_kwlist;
  Lilid *ptr_curr;

  if ((sql_command = malloc(sql_cmd_len)) == NULL) {
    LOG_PRINT(LOG_WARNING, get_status_msg(801));
    return 1;
  }

  /* connect to database server*/
  if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) == NULL) {
    LOG_PRINT(LOG_ERR, get_status_msg(205));
    free(sql_command);
    return 1;
  }

  /* get a list of all existing keywords */
  strcpy(sql_command, "SELECT * FROM t_keyword");
  LOG_PRINT(LOG_DEBUG, sql_command);

  dbires_kwlist = dbi_conn_query(conn, sql_command);
  if (!dbires_kwlist) {
    LOG_PRINT(LOG_WARNING, "keyword query error");
    free(sql_command);
    dbi_conn_close(conn);
    return 1;
  }

  if (ptr_sentinel) { /* we have a list of IDs to scan */
    if (!mode) {
      strcpy(sql_command, "SELECT refdb_id,refdb_abstract,refdb_title,refdb_booktitle,refdb_title_series FROM t_refdb WHERE refdb_id IN (");
    }
    else {
      strcpy(sql_command, "SELECT note_id,note_title,note_content FROM t_note WHERE note_id IN (");
    }

    ptr_curr = ptr_sentinel;

    /* loop over all references to scan */
    while ((ptr_curr = get_next_lilid(ptr_curr)) != NULL) {
      sprintf(id_buffer, ULLSPEC",", (unsigned long long)(ptr_curr->value));

      if ((new_sql_command = mstrcat(sql_command, id_buffer, &sql_cmd_len, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, get_status_msg(801));
	free(sql_command);
	dbi_result_free(dbires_kwlist);
	dbi_conn_close(conn);
	return 1;
      }
      else {
	sql_command = new_sql_command;
      }
    }
    if (*id_buffer) {
      /* overwrite trailing comma with closing bracket */
      if ((new_sql_command = mstrcat(sql_command, ")", &sql_cmd_len, 1)) == NULL) {
	LOG_PRINT(LOG_WARNING, get_status_msg(801));
	free(sql_command);
	dbi_result_free(dbires_kwlist);
	dbi_conn_close(conn);
	return 1;
      }
      else {
	sql_command = new_sql_command;
      }
    }
    else {
      /* no IDs found */
      free(sql_command);
      LOG_PRINT(LOG_INFO, "no IDs found for keyword scan");
      dbi_result_free(dbires_kwlist);
      dbi_conn_close(conn);
      return 0;
    }
  }
  else { /* scan all IDs in the database */
    if (!mode) {
      strcpy(sql_command, "SELECT refdb_id,refdb_abstract,refdb_title,refdb_booktitle,refdb_title_series FROM t_refdb WHERE refdb_id>0");
    }
    else {
      strcpy(sql_command, "SELECT note_id,note_title,note_content FROM t_note WHERE note_id>0");
    }
  }    

  LOG_PRINT(LOG_DEBUG, sql_command);

  dbires = dbi_conn_query(conn, sql_command);
  if (!dbires) {
    LOG_PRINT(LOG_WARNING, "reference query error");
    free(sql_command);
    dbi_result_free(dbires_kwlist);
    dbi_conn_close(conn);
    return 1;
  }

  /* loop over all IDs in our result set */
  while (dbi_result_next_row(dbires)) { 
    xref_id = my_dbi_result_get_idval(dbires, (mode) ? "note_id":"refdb_id");

    /* rewind keyword list */
    dbi_result_first_row(dbires_kwlist);

    /* loop over all keywords */
    while (dbi_result_next_row(dbires_kwlist)) {
      item_kw = my_dbi_result_get_string(dbires_kwlist, "keyword_name");
      keyword_id = my_dbi_result_get_idval(dbires_kwlist, "keyword_id");
      if (item_kw && *item_kw) {
	/* loop over fields to scan */
	i = 0;
	while (scanfields[mode][i][0]) {
	  item = my_dbi_result_get_string(dbires, scanfields[mode][i]);
	  if (item && *item && strstr(item, item_kw)) {
	    /* check for existing entry in t_xkeyword */
	    sprintf(sql_command, "SELECT keyword_id FROM t_xkeyword WHERE keyword_id="ULLSPEC" AND xkeyword_type=\'%s\' AND xref_id="ULLSPEC, (unsigned long long)keyword_id, (mode) ? "NOTE":"REFERENCE", (unsigned long long)xref_id);
	    LOG_PRINT(LOG_DEBUG, sql_command);
	    
	    dbires1 = dbi_conn_query(conn, sql_command);
	    if (!dbires1) {
	      LOG_PRINT(LOG_WARNING, "select error");
	      free(sql_command);
	      dbi_result_free(dbires);
	      dbi_result_free(dbires_kwlist);
	      dbi_conn_close(conn);
	      return 1;
	    }
	    if (!dbi_result_next_row(dbires1)) {
	      sprintf(sql_command, "INSERT INTO t_xkeyword (keyword_id,xref_id,xkeyword_type) VALUES ("ULLSPEC","ULLSPEC",\'%s\')", (unsigned long long)keyword_id, (unsigned long long)xref_id, (mode)? "NOTE":"REFERENCE");
	      LOG_PRINT(LOG_DEBUG, sql_command);
	      
	      dbires2 = dbi_conn_query(conn, sql_command);
	      if (!dbires2) {
		LOG_PRINT(LOG_WARNING, "insert error");
		free(sql_command);
		dbi_result_free(dbires);
		dbi_result_free(dbires_kwlist);
		dbi_conn_close(conn);
		return 1;
	      }
	      *ptr_insert++;

	      dbi_result_free(dbires2);
	      
	      if (!mode) {
		sprintf(sql_command, "noticed keyword %s in reference %s:"ULLSPEC, item_kw, ptr_clrequest->current_db, (unsigned long long)xref_id);
	      }
	      else {
		sprintf(sql_command, "noticed keyword %s in note %s:"ULLSPEC, item_kw, ptr_clrequest->current_db, (unsigned long long)xref_id);
	      }

	      LOG_PRINT(LOG_INFO, sql_command);

	      /* jump out as we don't have to scan further fields - the
		 keyword has been added anyway */
	      dbi_result_free(dbires1);
	      break;
	    }
	    dbi_result_free(dbires1);
	  }
	  i++;
	} /* end loop over all fields */
      } /* end if keyword */
    } /* end loop over all keywords */
  }
  dbi_result_free(dbires);

  free(sql_command);

  dbi_result_free(dbires_kwlist);
  dbi_conn_close(conn);

  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  deleteref(): implements the client command deleteref

  int deleteref returns >0 if failed, 0 if successful

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  struct ADDRESULT* ptr_addresult structure to hold number of successful and
                              failed deleterefs

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int deleteref(struct CLIENT_REQUEST* ptr_clrequest, struct ADDRESULT* ptr_addresult) {
  dbi_conn conn;
  int i;
  int numbyte; /* number of bytes written */
  int error;
  int retval = 0;
  int cs_status;
  size_t n_bufsize;
  char buffer[64];
  char* id_list;
  char* new_msg;
  const char* drivername;
  struct lilimem sentinel;
  Lilid id_sentinel;
  Lilid *ptr_curr;

  sentinel.ptr_mem = NULL;
  sentinel.ptr_next = NULL;
  sentinel.varname[0] = '\0';

  id_sentinel.ptr_next = NULL;

  n_bufsize = atoi(ptr_clrequest->argument);

  /* try to allocate the amount the client requested */
  id_list = malloc(n_bufsize);
  if (id_list == NULL || insert_lilimem(&sentinel, (void**)&id_list, NULL)) {
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 1;
  }

  ptr_addresult->msg = malloc(256); /* something to start with */

  if (!ptr_addresult->msg) {
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 1;
  }

  *(ptr_addresult->msg) = '\0';
  ptr_addresult->msg_len = 256;

  /* send acknowledgement to client */
  send_status(ptr_clrequest->fd, 0, TERM_NO);

  if ((cs_status = read_status(ptr_clrequest->fd)) != 0) {
    LOG_PRINT(LOG_INFO, get_status_msg(112));
    delete_all_lilimem(&sentinel);
    return 1;
  }

  /* read id list from client */
  numbyte = tread(ptr_clrequest->fd, id_list, n_bufsize);
  if (numbyte == -1) {
    LOG_PRINT(LOG_INFO, get_status_msg(109));
    delete_all_lilimem(&sentinel);
    return 1;
  }
  
  /* split id_list into tokens */
  error = string_tokenize_lili(&id_sentinel, id_list);
  if (error == 2) {
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    delete_all_lilid(&id_sentinel);
    delete_all_lilimem(&sentinel);
    return 1;
  }
  else if (error == 1) {
    send_status(ptr_clrequest->fd, 412, TERM_NO);
    delete_all_lilid(&id_sentinel);
    delete_all_lilimem(&sentinel);
    return 1;
  }

  /* connect to database server*/
  if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) == NULL) {
    send_status(ptr_clrequest->fd, 204, TERM_NO);
    LOG_PRINT(LOG_WARNING, get_status_msg(204));
    delete_all_lilid(&id_sentinel);
    delete_all_lilimem(&sentinel);
    return 1;
  }

  drivername = dbi_driver_get_name(dbi_conn_get_driver(conn));


  ptr_curr = &id_sentinel;

  /* loop over arguments */
  while ((ptr_curr = get_next_lilid(ptr_curr)) != NULL) {
    if ((retval = delete_ref_by_id(ptr_curr->value, conn, ptr_clrequest, ptr_addresult)) != 0) {
      sprintf(buffer, "420:"ULLSPEC"\n", (unsigned long long)ptr_curr->value);
      if ((new_msg = mstrcat(ptr_addresult->msg, buffer, &(ptr_addresult->msg_len), 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return 1;
      }
      else {
	ptr_addresult->msg = new_msg;
      }
    }
  } /* end while */

  if (conn && ptr_addresult->success) {
    update_meta(conn, ptr_clrequest);
  }

  dbi_conn_close(conn);
  delete_all_lilid(&id_sentinel);
  delete_all_lilimem(&sentinel);
  
  send_status(ptr_clrequest->fd, 0, TERM_NO);

  numbyte = tiwrite(ptr_clrequest->fd, ptr_addresult->msg, TERM_YES);
  if (numbyte == -1) {
    LOG_PRINT(LOG_INFO, get_status_msg(110));
    retval = 1;
  }

  return retval;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  delete_ref_by_id(): deletes a reference by its ID value

  int delete_ref_by_id returns 0 if failed, 1 if successful

  unsigned long long idval ID value of the note to be deleted

  dbi_conn conn database connection

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  struct ADDRESULT* ptr_addresult structure to hold number of successful and
                              failed deleterefs

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int delete_ref_by_id(unsigned long long idval, dbi_conn conn, struct CLIENT_REQUEST* ptr_clrequest, struct ADDRESULT* ptr_addresult) {
  char sql_command[640]; /* fixed length is ok here since only IDs are appended to the query strings, and the length of an ID is limited by the database */
  char periodical_id[32] = ""; /* will hold the ID until used */
  char* new_msg;
  int result;
  unsigned long long n_id = 0;
  unsigned long long n_xid = 0;
  unsigned long long n_periodical_id;
  unsigned long long numrows;
  dbi_result dbires;
  dbi_result dbires1;
  dbi_result dbires2;
  dbi_result dbires_user;

  /* The general procedure to remove a reference is as follows: Remove
     the entry in the main table t_refdb. Query the author, keyword, user
     crosslink tables for authors, keywords, users used by this reference.
     Find out whether these authors, keywords, users are used by any other
     datasets. If not, delete them from the authors, keywords, users data
     tables. In both cases, delete the entries in the crosslink tables */
    
  /* If the db server supports it, start a transaction. We want one
     transaction per reference */
  if (my_dbi_conn_begin(conn)) {
    LOG_PRINT(LOG_WARNING, get_status_msg(227));
    return 227;
  }
  
  /* lock the tables we'll write to to prevent concurrent writes
     from different clients */
  if (my_dbi_conn_lock(conn)) {
    LOG_PRINT(LOG_WARNING, get_status_msg(228));
    return 228;
  }

  sprintf(sql_command, "SELECT refdb_id FROM t_refdb WHERE refdb_type != \'DUMMY\' AND refdb_id="ULLSPEC, (unsigned long long)idval);
  LOG_PRINT(LOG_DEBUG, sql_command);
  dbires = dbi_conn_query(conn, sql_command);
  if (!dbires) {
    my_dbi_conn_unlock(conn);
    my_dbi_conn_rollback(conn);
    LOG_PRINT(LOG_WARNING, get_status_msg(234));
    return 234;
  }
    
  if (dbi_result_get_numrows(dbires) == 0) {
    /* ID does not exist - nothing to do */
    ptr_addresult->skipped++;
    dbi_result_free(dbires);
    my_dbi_conn_unlock(conn);
    my_dbi_conn_commit(conn);

    /* add message, reuse sql_command */
    sprintf(sql_command, "417:"ULLSPEC"\n", (unsigned long long)idval);
    if ((new_msg = mstrcat(ptr_addresult->msg, sql_command, &(ptr_addresult->msg_len), 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      ptr_addresult->msg = new_msg;
    }
    return 417;
  }
    
  dbi_result_free(dbires);

  sprintf(sql_command, "SELECT t_user.user_id, t_user.user_name, t_xuser.xuser_id FROM t_user, t_xuser WHERE t_user.user_id=t_xuser.user_id AND t_xuser.refdb_id="ULLSPEC, (unsigned long long)idval);
  LOG_PRINT(LOG_DEBUG, sql_command);
  dbires_user = dbi_conn_query(conn, sql_command);
  if (!dbires_user) {
    my_dbi_conn_unlock(conn);
    my_dbi_conn_rollback(conn);
    LOG_PRINT(LOG_WARNING, get_status_msg(234));
    return 234;
  }

  if (dbi_result_get_numrows(dbires_user) >= 1) {
    int used_by_other = 1;

    if (dbi_result_next_row(dbires_user) == 0) {
      dbi_result_free(dbires_user);
      my_dbi_conn_unlock(conn);
      my_dbi_conn_rollback(conn);
      LOG_PRINT(LOG_WARNING, get_status_msg(234));
      return 234;
    }

    /* save for later use */
    n_id = my_dbi_result_get_idval(dbires_user, "user_id");
    n_xid = my_dbi_result_get_idval(dbires_user, "xuser_id");

    if (dbi_result_get_numrows(dbires_user) == 1) {

      if (!strcmp(ptr_clrequest->username, my_dbi_result_get_string(dbires_user, "user_name"))) {
	used_by_other--;
      }
    }

    if (used_by_other) {
      /* dataset is used by at least one other user, refuse to delete it */
      ptr_addresult->failure++;


      dbi_result_free(dbires_user);
      my_dbi_conn_unlock(conn);
      my_dbi_conn_rollback(conn);

      /* send message to client, reuse sql_command */
      sprintf(sql_command, "423:"ULLSPEC"\n", (unsigned long long)idval);
      if ((new_msg = mstrcat(ptr_addresult->msg, sql_command, &(ptr_addresult->msg_len), 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return 801;
      }
      else {
	ptr_addresult->msg = new_msg;
      }

      return 423;
    }
  }

  /* retrieve periodical_id for later use */
  sprintf(sql_command, "SELECT refdb_periodical_id FROM t_refdb WHERE refdb_id="ULLSPEC, (unsigned long long)idval);
  LOG_PRINT(LOG_DEBUG, sql_command);

  dbires = dbi_conn_query(conn, sql_command);
  if (!dbires) {
    dbi_result_free(dbires_user);
    my_dbi_conn_unlock(conn);
    my_dbi_conn_rollback(conn);
    LOG_PRINT(LOG_WARNING, get_status_msg(234));
    return 234;
  }

  if (dbi_result_next_row(dbires) != 0) {
    if ((n_periodical_id = my_dbi_result_get_idval(dbires, "refdb_periodical_id")) != 0) {
      sprintf(periodical_id, ULLSPEC, (unsigned long long)n_periodical_id); /* save ID */
    }
  }

  dbi_result_free(dbires);

  /* search orphans in t_keyword */
  result = remove_keyword_entries(idval, conn, 0);

  if (result) {
    /* send message to client, reuse sql_command */
    sprintf(sql_command, "229:"ULLSPEC"\n", (unsigned long long)idval);
    if ((new_msg = mstrcat(ptr_addresult->msg, sql_command, &(ptr_addresult->msg_len), 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      ptr_addresult->msg = new_msg;
    }
    dbi_result_free(dbires_user);
    my_dbi_conn_unlock(conn);
    my_dbi_conn_rollback(conn);
    return 229;
  }

  /* search orphans in t_author */
  result = remove_author_entries(idval, conn);

  if (result) {
    /* send message to client, reuse sql_command */
    sprintf(sql_command, "230:"ULLSPEC"\n", (unsigned long long)idval);
    if ((new_msg = mstrcat(ptr_addresult->msg, sql_command, &(ptr_addresult->msg_len), 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      ptr_addresult->msg = new_msg;
    }
    dbi_result_free(dbires_user);
    my_dbi_conn_unlock(conn);
    my_dbi_conn_rollback(conn);
    return 230;
  }


  /* search orphans in t_link */
  result = remove_ulink_entries(idval, conn, 0 /* reference */);

  if (result) {
    /* send message to client, reuse sql_command */
    sprintf(sql_command, "259:"ULLSPEC"\n", (unsigned long long)idval);
    if ((new_msg = mstrcat(ptr_addresult->msg, sql_command, &(ptr_addresult->msg_len), 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      ptr_addresult->msg = new_msg;
    }
    dbi_result_free(dbires_user);
    my_dbi_conn_unlock(conn);
    my_dbi_conn_rollback(conn);
    return 259;
  }

  /* remove orphans in t_xnote */
  result = remove_xnote_entries(idval, conn, 0 /*ref*/);

  if (result != 0 && result != 4) {
    /* send message to client, reuse sql_command */
    sprintf(sql_command, "245:"ULLSPEC"\n", (unsigned long long)idval);
    if ((new_msg = mstrcat(ptr_addresult->msg, sql_command, &(ptr_addresult->msg_len), 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      ptr_addresult->msg = new_msg;
    }
    dbi_result_free(dbires_user);
    my_dbi_conn_unlock(conn);
    my_dbi_conn_rollback(conn);
    return 245;
  }


  /* search orphans in t_periodical */
  if (n_periodical_id) {
    result = remove_periodical_entries(n_periodical_id, conn);

    if (result) {
    /* send message to client, reuse sql_command */
    sprintf(sql_command, "231:"ULLSPEC"\n", (unsigned long long)idval);
    if ((new_msg = mstrcat(ptr_addresult->msg, sql_command, &(ptr_addresult->msg_len), 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      ptr_addresult->msg = new_msg;
    }
    dbi_result_free(dbires_user);
    my_dbi_conn_unlock(conn);
    my_dbi_conn_rollback(conn);
    return 231;
    }
  }

  /* delete entry in main table */
  sprintf(sql_command, "DELETE FROM t_refdb WHERE refdb_id="ULLSPEC, idval);
  LOG_PRINT(LOG_DEBUG, sql_command);
  dbires = dbi_conn_query(conn, sql_command);
  if (!dbires) {
    /* send message to client, reuse sql_command */
    sprintf(sql_command, "248:"ULLSPEC"\n", (unsigned long long)idval);
    if ((new_msg = mstrcat(ptr_addresult->msg, sql_command, &(ptr_addresult->msg_len), 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      ptr_addresult->msg = new_msg;
    }
    dbi_result_free(dbires_user);
    my_dbi_conn_unlock(conn);
    my_dbi_conn_rollback(conn);
    LOG_PRINT(LOG_WARNING, get_status_msg(248));
    return 248;
  }

  dbi_result_free(dbires);

  /* search orphans in t_user. The query returning dbires_user ran
     above. The loop should run only once */
  /* delete entry in xuser table */
  sprintf(sql_command, "DELETE FROM t_xuser WHERE xuser_id="ULLSPEC, (unsigned long long)n_xid);
  LOG_PRINT(LOG_DEBUG, sql_command);
  dbires = dbi_conn_query(conn, sql_command);
  if (!dbires) {
    /* send message to client, reuse sql_command */
    sprintf(sql_command, "234:"ULLSPEC"\n", (unsigned long long)idval);
    if ((new_msg = mstrcat(ptr_addresult->msg, sql_command, &(ptr_addresult->msg_len), 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      ptr_addresult->msg = new_msg;
    }
    dbi_result_free(dbires_user);
    my_dbi_conn_unlock(conn);
    my_dbi_conn_rollback(conn);
    LOG_PRINT(LOG_WARNING, get_status_msg(234));
    return 234;
  }

  dbi_result_free(dbires);

  /* delete entry in user table */
  result = remove_user_entries(n_id, conn);
  
  if (result != 0 && result != 4) {
    /* send message to client, reuse sql_command */
    sprintf(sql_command, "234:"ULLSPEC"\n", (unsigned long long)idval);
    if ((new_msg = mstrcat(ptr_addresult->msg, sql_command, &(ptr_addresult->msg_len), 0)) == NULL) {
      dbi_result_free(dbires_user);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      ptr_addresult->msg = new_msg;
    }
    dbi_result_free(dbires_user);
    my_dbi_conn_unlock(conn);
    my_dbi_conn_rollback(conn);
    LOG_PRINT(LOG_WARNING, get_status_msg(234));
    return 234;
  }

  dbi_result_free(dbires_user);
  ptr_addresult->success++;

  /* send message to client, reuse sql_command */
  sprintf(sql_command, "419:"ULLSPEC"\n", (unsigned long long)idval);
  if ((new_msg = mstrcat(ptr_addresult->msg, sql_command, &(ptr_addresult->msg_len), 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 801;
  }
  else {
    ptr_addresult->msg = new_msg;
  }

  my_dbi_conn_unlock(conn);
  my_dbi_conn_commit(conn);

  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  pickref(): implements the client command pickref

  int pickref returns >0 if failed, 0 if successful

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  int n_remove if 1, remove from personal interest list. if 0, add to list

  struct ADDRESULT* ptr_addresult structure to hold number of successful and
                              failed pickrefs

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int pickref(struct CLIENT_REQUEST* ptr_clrequest, int n_remove, struct ADDRESULT* ptr_addresult) {
  dbi_conn conn;
  dbi_result dbires;
  dbi_result dbires1;
  int error = 0;
  int numrows;
  int insert_len;
  int i;
  int numbyte; /* number of bytes written */
  int n_bufsize;
  int cs_status;
  int retval = 0;
  unsigned long long n_id;
  unsigned long long num_entries;
  unsigned long long n_user_id;
  size_t sql_cmd_len; /* length of the sql_command buffer */
  char the_id[32];
  char *sql_command; /* a buffer for the sql commands we got to piece together */
  char *new_sql_command; /* for realloc()ing sql_command */
  char *new_msg;
  char *ignore_entry; /* a list of IDs that will not be added because they're duplicates */
  char *exist_entry; /* a list of IDs that will be added as they exist */
  char *id_list; /* the list of IDs that the client sends in */
  char return_msg[64]; /* string for a message to the client */
  const char *drivername; /* name of the libdbi driver */
  struct lilimem sentinel;
  Lilid id_sentinel;
  Lilid *ptr_curr;

  /* initialize linked lists */
  sentinel.ptr_mem = NULL;
  sentinel.ptr_next = NULL;
  sentinel.varname[0] = '\0';

  id_sentinel.ptr_next = NULL;

  /* allocate memory */
  sql_cmd_len = 256;
  sql_command = malloc(sql_cmd_len);
  if (sql_command == NULL || insert_lilimem(&sentinel, (void**)&sql_command, NULL)) {
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 1;
  }

  ptr_addresult->msg_len = 512;
  ptr_addresult->msg = malloc(ptr_addresult->msg_len);
  if (ptr_addresult->msg == NULL) {
    free(sql_command);
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 1;
  }

  *(ptr_addresult->msg) = '\0';

  n_bufsize = atoi(ptr_clrequest->argument);

  /* try to allocate the amount the client requested */
  id_list = malloc(n_bufsize);
  if (id_list == NULL || insert_lilimem(&sentinel, (void**)&id_list, NULL)) {
    delete_all_lilimem(&sentinel);
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 1;
  }

  /* send acknowledgement to client */
  send_status(ptr_clrequest->fd, 0, TERM_NO);

  /* read id list from client */
  if ((cs_status = read_status(ptr_clrequest->fd)) != 0) {
    LOG_PRINT(LOG_INFO, get_status_msg(112));
    delete_all_lilimem(&sentinel);
    return 1;
  }

  numbyte = tread(ptr_clrequest->fd, id_list, n_bufsize);
  if (numbyte == -1) {
    delete_all_lilimem(&sentinel);
    LOG_PRINT(LOG_ERR, get_status_msg(109));
    return 1;
  }
  
  /* split id_list into tokens */
  error = string_tokenize_lili(&id_sentinel, id_list);
  if (error == 2) {
    delete_all_lilimem(&sentinel);
    delete_all_lilid(&id_sentinel);
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 1;
  }
  else if (error == 1) {
    delete_all_lilimem(&sentinel);
    delete_all_lilid(&id_sentinel);
    send_status(ptr_clrequest->fd, 412, TERM_NO);
    LOG_PRINT(LOG_INFO, get_status_msg(412));
    return 2;
  }

  /* connect to database server*/
  if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) == NULL) {
    delete_all_lilid(&id_sentinel);
    delete_all_lilimem(&sentinel);
    send_status(ptr_clrequest->fd, 204, TERM_NO);
    LOG_PRINT(LOG_WARNING, get_status_msg(204));
    return 1;
  }

  /* The general procedure to pick a reference is:
     check whether there is an entry in t_xuser for the current
     user pointing to this refdb_id. If yes and we shall add,
     we're done. If yes and we shall remove, do so. If no entry is
     found: If we shall remove, we're done, if we shall add, do so.
     We also have to add the user to t_user if he's not already
     there. And upon removing entries from t_xuser we have to check
     whether we remove the last reference to the user in t_user. If
     so, we have to kick him out */

  /* If the db server supports it, start a transaction. We want one
     transaction per reference */
  if (my_dbi_conn_begin(conn)) {
    send_status(ptr_clrequest->fd, 227, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(227));
    retval = 1;
    goto Finish;
  }
  
  /* lock the tables we'll write to to prevent concurrent writes
     from different clients */
  if (my_dbi_conn_lock(conn)) {
    send_status(ptr_clrequest->fd, 228, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(228));
    retval = 1;
    goto Finish;
  }

  /* get the user_id for the current user */
  sprintf(sql_command, "SELECT user_id FROM t_user WHERE user_name=\'%s\'", ptr_clrequest->username);
  
  LOG_PRINT(LOG_DEBUG, sql_command);

  dbires = dbi_conn_query(conn, sql_command);
  if (!dbires) {
    send_status(ptr_clrequest->fd, 234, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(234));
    my_dbi_conn_unlock(conn);
    my_dbi_conn_rollback(conn);
    retval = 1;
    goto Finish;
  }

  if (dbi_result_get_numrows(dbires) == 0) {
    if (!n_remove) {
      /* create an entry for the current user in t_user */
      sprintf(sql_command, "INSERT INTO t_user (user_name) VALUES (\'%s\')", ptr_clrequest->username);

      LOG_PRINT(LOG_DEBUG, sql_command);

      dbires1 = dbi_conn_query(conn, sql_command);
      if (!dbires1) {
	send_status(ptr_clrequest->fd, 233, TERM_NO);
	LOG_PRINT(LOG_CRIT, get_status_msg(233));
	dbi_result_free(dbires);
	my_dbi_conn_unlock(conn);
	my_dbi_conn_rollback(conn);
	retval = 1;
	goto Finish;
      }

      dbi_result_free(dbires1);

      /* retrieve the generated user_id in t_user */
      drivername = dbi_driver_get_name(dbi_conn_get_driver(conn));

      if (!strcmp(my_dbi_conn_get_cap(conn, "named_seq"), "f")) {
	n_user_id = dbi_conn_sequence_last(conn, NULL);
      }
      else {
	n_user_id = dbi_conn_sequence_last(conn, "t_user_user_id_seq");
      }
    }
  }
  else { /* numrows > 0 */

    /* retrieve the user_id from t_user */
    if (dbi_result_next_row(dbires) != 0) {
      n_user_id = (unsigned long long)my_dbi_result_get_idval(dbires, "user_id");
    }
    else {
      send_status(ptr_clrequest->fd, 234, TERM_NO);
      LOG_PRINT(LOG_WARNING, get_status_msg(234));
      dbi_result_free(dbires);
      my_dbi_conn_unlock(conn);
      my_dbi_conn_rollback(conn);
      retval = 1;
      goto Finish;
    }
  }

  dbi_result_free(dbires);

  /*----------------------------------------------------------------*/
  /* remove entries */
  if (n_remove) {
    /* remove entries one by one to get exact status reports */
    ptr_curr = &id_sentinel;

    /* loop over arguments */
    while ((ptr_curr = get_next_lilid(ptr_curr)) != NULL) {
      sprintf(sql_command, "DELETE FROM t_xuser WHERE refdb_id="ULLSPEC" AND user_id="ULLSPEC, (unsigned long long)(ptr_curr->value), (unsigned long long)n_user_id);
      LOG_PRINT(LOG_DEBUG, sql_command);

      dbires = dbi_conn_query(conn, sql_command);
      if (!dbires) {
	/* reuse sql_command */
	sprintf(sql_command, "253:"ULLSPEC":%s\n", (unsigned long long)(ptr_curr->value), ptr_clrequest->username);
	ptr_addresult->failure++;
      }
      else {
	if (dbi_result_get_numrows_affected(dbires)) {
	  sprintf(sql_command, "254:"ULLSPEC":%s\n", (unsigned long long)(ptr_curr->value), ptr_clrequest->username);
	  ptr_addresult->success++;
	}
	else {
	  sprintf(sql_command, "417:"ULLSPEC":%s\n", (unsigned long long)(ptr_curr->value), ptr_clrequest->username);
	  ptr_addresult->skipped++;
	}

	dbi_result_free(dbires);
      }

      if ((new_msg = mstrcat(ptr_addresult->msg, sql_command, &(ptr_addresult->msg_len), 0)) == NULL) {
	send_status(ptr_clrequest->fd, 801, TERM_NO);
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	retval = 1;
	goto Finish;
      }
      else {
	ptr_addresult->msg = new_msg;
      }
    }

 /* search orphans in t_user */
    sprintf(sql_command, "SELECT xuser_id FROM t_xuser WHERE user_id="ULLSPEC, (unsigned long long)n_user_id);

    LOG_PRINT(LOG_DEBUG, sql_command);

    dbires = dbi_conn_query(conn, sql_command);
    if (!dbires) {
      send_status(ptr_clrequest->fd, 234, TERM_NO);
      LOG_PRINT(LOG_WARNING, get_status_msg(234));
      my_dbi_conn_unlock(conn);
      my_dbi_conn_rollback(conn);
      retval = 1;
      goto Finish;
    }

    numrows = dbi_result_get_numrows(dbires);

    if (!numrows) { /* if user picked no other references */
	/* delete entry in t_user */
      sprintf(sql_command, "DELETE FROM t_user WHERE user_id="ULLSPEC, (unsigned long long)n_user_id);

      LOG_PRINT(LOG_DEBUG, sql_command);

      dbires1 = dbi_conn_query(conn, sql_command);
      if (!dbires1) {
	send_status(ptr_clrequest->fd, 247, TERM_NO);
	LOG_PRINT(LOG_WARNING, get_status_msg(247));
	dbi_result_free(dbires);
	my_dbi_conn_unlock(conn);
	my_dbi_conn_rollback(conn);
	retval = 1;
	goto Finish;
      }
    }
    dbi_result_free(dbires);
  }

  /*----------------------------------------------------------------*/
  /* add entries */
  else { /* if (!n_remove) */
    /* add entries one by one to get exact status reports */
    ptr_curr = &id_sentinel;

    /* loop over arguments */
    while ((ptr_curr = get_next_lilid(ptr_curr)) != NULL) {

      /* make sure the target exists */
      sprintf(sql_command, "SELECT refdb_id FROM t_refdb WHERE refdb_id="ULLSPEC, (unsigned long long)(ptr_curr->value));
      LOG_PRINT(LOG_DEBUG, sql_command);

      dbires = dbi_conn_query(conn, sql_command);
      if (!dbires) {
	send_status(ptr_clrequest->fd, 234, TERM_NO);
	LOG_PRINT(LOG_WARNING, get_status_msg(234));
	my_dbi_conn_unlock(conn);
	my_dbi_conn_rollback(conn);
	retval = 1;
	goto Finish;
      }

      if (!dbi_result_get_numrows(dbires)) {
	/* target ID missing */
	dbi_result_free(dbires);
	/* reuse sql_command */
	sprintf(sql_command, "412:"ULLSPEC"\n", (unsigned long long)(ptr_curr->value));
	ptr_addresult->skipped++;
	if ((new_msg = mstrcat(ptr_addresult->msg, sql_command, &(ptr_addresult->msg_len), 0)) == NULL) {
	  send_status(ptr_clrequest->fd, 801, TERM_NO);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  retval = 1;
	  goto Finish;
	}
	else {
	  ptr_addresult->msg = new_msg;
	}
	continue;
      }

      dbi_result_free(dbires);

      /* check for existing entry in t_xuser */
      sprintf(sql_command, "SELECT xuser_id FROM t_xuser WHERE user_id="ULLSPEC" AND refdb_id="ULLSPEC, (unsigned long long)n_user_id, (unsigned long long)(ptr_curr->value));

      LOG_PRINT(LOG_DEBUG, sql_command);

      dbires = dbi_conn_query(conn, sql_command);
      if (!dbires) {
	send_status(ptr_clrequest->fd, 234, TERM_NO);
	LOG_PRINT(LOG_WARNING, get_status_msg(234));
	my_dbi_conn_unlock(conn);
	my_dbi_conn_rollback(conn);
	retval = 1;
	goto Finish;
      }

      if (dbi_result_get_numrows(dbires)) {
	/* link already exists */
	dbi_result_free(dbires);
	/* reuse sql_command */
	sprintf(sql_command, "424:"ULLSPEC":%s\n", (unsigned long long)(ptr_curr->value), ptr_clrequest->username);
	ptr_addresult->skipped++;
	if ((new_msg = mstrcat(ptr_addresult->msg, sql_command, &(ptr_addresult->msg_len), 0)) == NULL) {
	  send_status(ptr_clrequest->fd, 801, TERM_NO);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  retval = 1;
	  goto Finish;
	}
	else {
	  ptr_addresult->msg = new_msg;
	}
	continue;
      }

      dbi_result_free(dbires);

      /* insert new link */
      sprintf(sql_command, "INSERT INTO t_xuser (xuser_reprint, user_id, refdb_id) VALUES ('NOT IN FILE',"ULLSPEC","ULLSPEC")", (unsigned long long)n_user_id, (unsigned long long)(ptr_curr->value));
      LOG_PRINT(LOG_DEBUG, sql_command);

      dbires = dbi_conn_query(conn, sql_command);
      if (!dbires) {
	/* reuse sql_command */
	sprintf(sql_command, "255:"ULLSPEC":%s\n", (unsigned long long)(ptr_curr->value), ptr_clrequest->username);
	ptr_addresult->failure++;
      }
      else {
	if (dbi_result_get_numrows_affected(dbires)) {
	  sprintf(sql_command, "256:"ULLSPEC":%s\n", (unsigned long long)(ptr_curr->value), ptr_clrequest->username);
	  ptr_addresult->success++;
	}
	else {
	  sprintf(sql_command, "407:"ULLSPEC":%s\n", (unsigned long long)(ptr_curr->value), ptr_clrequest->username);
	  ptr_addresult->skipped++;
	}

	dbi_result_free(dbires);
      }

      if ((new_msg = mstrcat(ptr_addresult->msg, sql_command, &(ptr_addresult->msg_len), 0)) == NULL) {
	send_status(ptr_clrequest->fd, 801, TERM_NO);
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	retval = 1;
	goto Finish;
      }
      else {
	ptr_addresult->msg = new_msg;
      }
    }
  }

  my_dbi_conn_unlock(conn);
  my_dbi_conn_commit(conn);

  /* send report to client */
  send_status(ptr_clrequest->fd, 0, TERM_NO);

  numbyte = tiwrite(ptr_clrequest->fd, ptr_addresult->msg, TERM_YES);
  if (numbyte == -1) {
    LOG_PRINT(LOG_INFO, get_status_msg(110));
    retval = 1;
  }


 Finish:   
  if (conn && ptr_addresult->success) {
    update_meta(conn, ptr_clrequest);
  }

  dbi_conn_close(conn);
  delete_all_lilid(&id_sentinel);
  delete_all_lilimem(&sentinel);
  return retval;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  updatejo(): implements the client command updatejo

  int updatejo returns 1 if failed, 0 if successful

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  struct ADDRESULT* addresult this structure will be filled in with the number
                      of (un-)successfully added/updated references

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int updatejo(struct CLIENT_REQUEST* ptr_clrequest, struct ADDRESULT* ptr_addresult) {
  dbi_conn conn;
  dbi_result dbires;
  dbi_driver driver;
  char **inargv; /* tokens of the client request */
  int inargc; /* number of tokens of the client request */
  int inargcmax; /* maximum number of tokens of the client request */
  int len;
  int result;
  int mode;
  int retval = 0;
  int havefirst = 0;
  int error = 0; /* codes: 0 = no error, 1 = out of memory, 2 = subselect failed */
  int syn_count = 0; /* counter for synonyms changed */
  unsigned long long n_id;
  unsigned long long n_target_id;
  size_t buffer_len; /* these are the allocated lengths of the buffers */
  size_t sql_command_len;
  size_t return_msg_len;
  char *return_msg;
  char *sql_command; /* these are ptrs to the buffers and temporary ptrs */
  char *new_sql_command;
  char *buffer;
  char *new_buffer;
  char *token;
  char *newtoken;
  char *stripped_token;
  char *quoted_token;
  char *eostring;
  char journal_val[256] = ""; /* name of journal*/
  char journal_field[23] = ""; /* t_periodical field name of synonym */
  char field_buffer[10];
  char my_type[20];
  char assemble_buffer[2048];
  char jo[256] = "";
  char jf[256] = "";
  char j1[256] = "";
  char j2[256] = "";
  struct SQLTOKEN sqltoken;
  struct lilimem sentinel;
  Liliform *ptr_curr;
  Liliform syn_sentinel;

  ptr_addresult->success = 0;
  ptr_addresult->failure = 0;
  ptr_addresult->updated = 0;
  ptr_addresult->skipped = 0;

  sentinel.ptr_mem = NULL;
  sentinel.ptr_next = NULL;
  sentinel.varname[0] = '\0';

  syn_sentinel.ptr_next = NULL;
  syn_sentinel.name[0] = '\0';
  syn_sentinel.value = NULL;

  /* get buffer to analyze the request */
  inargc = 0;
  inargcmax = 10;
  inargv = malloc((size_t)inargcmax*sizeof(char*));
  if (inargv == NULL || insert_lilimem(&sentinel, (void**)&inargv, NULL)) {
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 1;
  }

  /* get buffer to assemble the SQL queries */
  sql_command_len = 4096; 
  sql_command = malloc(sql_command_len); 

  if (sql_command == NULL || insert_lilimem(&sentinel, (void**)&sql_command, NULL)) {
    delete_all_lilimem(&sentinel);
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 1;
  }
  *sql_command = '\0'; /* start with an empty string */

  /* get a buffer to hold tokens */
  buffer_len = 4096;
  buffer = malloc(buffer_len); 

  if (buffer == NULL || insert_lilimem(&sentinel, (void**)&buffer, NULL)) {
    delete_all_lilimem(&sentinel);
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 1;
  }
  *buffer = '\0'; /* start with an empty string */

  /* yet another buffer */
  return_msg_len = 4096;
  return_msg = malloc(return_msg_len); 

  if (return_msg == NULL || insert_lilimem(&sentinel, (void**)&return_msg, NULL)) {
    delete_all_lilimem(&sentinel);
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 1;
  }
  *return_msg = '\0'; /* start with an empty string */

  eostring = ptr_clrequest->argument + strlen(ptr_clrequest->argument);

  /* connect to the database */
  if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) == NULL) {
    send_status(ptr_clrequest->fd, 204, TERM_NO);
    LOG_PRINT(LOG_WARNING, get_status_msg(204));
    delete_all_lilimem(&sentinel);
    return 1;
  }

  driver = dbi_conn_get_driver(conn);

  token = ptr_clrequest->argument;
  newtoken = token;

/*   printf("%s\n", token); */

  /* general strategy: scan the command string for items and add all of
     them except the first to a linked list. Then walk through
     the linked list to assemble the update query */


  /* loop as long as we find more tokens */
  while (newtoken != NULL) {
    token = link_tokenize(newtoken, &sqltoken);
    newtoken = sqltoken.next_token;
/*     printf("token:%s<<newtoken:%s<<\n", token, newtoken); */

    if (token != NULL) {
      /* extract the token and save to a temporary string */
      if (sqltoken.length > buffer_len) {
	new_buffer = (char*)realloc(buffer, (size_t)sqltoken.length);
	if (new_buffer == NULL) { /* out of memory */
	  error = 1;
	  break;
	}
	else {
	  buffer = new_buffer;
	  buffer_len = sqltoken.length;
	}
      }
      strncpy(buffer, token, sqltoken.length);
      buffer[sqltoken.length] ='\0';
      if (sqltoken.type == 4) { 
	/*----------------------------------------------------------------*/
	/* target fields */
	if (strncmp(token, ":JF:", 4) == 0 || /* journal full */
	    strncmp(token, ":JO:", 4) == 0 || /* journal abbrev */
	    strncmp(token, ":JA:", 4) == 0 || /* journal abbrev */
	    strncmp(token, ":J1:", 4) == 0 || /* journal custabbrev1 */
	    strncmp(token, ":J2:", 4) == 0) { /* journal custabbrev2 */

	  strcpy(field_buffer, buffer);
	  field_buffer[4] = '\0'; /* terminate string */

	  /* obtain the token after the '=' */
	  token = link_tokenize(newtoken, &sqltoken);
	  newtoken = sqltoken.next_token;
	  if (token != NULL) {
	    strncpy(buffer, token, sqltoken.length);
	    buffer[sqltoken.length] ='\0'; /* terminate string */
	    len = 0;

	    token = nstrtok(buffer, &len, " ");
	    if (token != NULL) {
/* 	      printf("token went to: %s<<\n", token); */
	      strncpy(buffer, token, sqltoken.length);
	      buffer[sqltoken.length] ='\0'; /* terminate string */
	      len = 0;
	      token = nstrtok(buffer, &len, " ");
	      if (token != NULL) {
		quoted_token = malloc(strlen(token)+1);
		if (!quoted_token) {
		  error = 1;
		  break;
		}

		stripped_token = strip_quote(token);

		unescape_chars(quoted_token, stripped_token, strlen(stripped_token));


/* 		printf("quoted_token went to: %s<<\n", quoted_token); */

		if (!havefirst) {
		  if (!strcmp(field_buffer, ":JF:")) {
		    strcpy(journal_field, "periodical_name");
		  }
		  else if (!strcmp(field_buffer, ":JO:")
			   || !strcmp(field_buffer, ":JA:")) {
		    strcpy(journal_field, "periodical_abbrev");
		  }
		  else if (!strcmp(field_buffer, ":J1:")) {
		    strcpy(journal_field, "periodical_custabbrev1");
		  }
		  else {
		    strcpy(journal_field, "periodical_custabbrev2");
		  }

		  if (dbi_conn_quote_string(conn, &quoted_token) == 0) {
		    error = 1;
		    break;
		  }

		  strncpy(journal_val, quoted_token, 255);
		  journal_val[255] = '\0';
		  /* 	      printf("key:%s<<value:%s<<\n", journal_type, journal_val); */
		  free(quoted_token);

		  if (is_journal(conn, journal_field, journal_val) != 1) {
		    error = 2;
		    break;
		  }
		  havefirst++;
		}
		else {
		  if (insert_liliform(&syn_sentinel, field_buffer, quoted_token)) {
		    error = 1;
		    break;
		  }
		}
	      }
	    }
	  }
	}
	else {
	  /* ignore unknown specifier */
	  ptr_addresult->skipped++;
	}
      } /* if (sqltoken.type != 4) / else */
    } /* if (token != NULL) */
  } /* while (newtoken != NULL) */

  if (error || !*journal_field || !*journal_val) {
    if (error == 1) {
      send_status(ptr_clrequest->fd, 801, TERM_NO);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
    }
    else {
      send_status(ptr_clrequest->fd, 417, TERM_NO);
      LOG_PRINT(LOG_INFO, get_status_msg(417));
    }
    retval = 1;
    goto Finish;
  }

  /* loop over all synonyms. If the same synonym is specified more
     than once, the last value prevails */
  ptr_curr = &syn_sentinel;

  while ((ptr_curr = get_next_liliform(ptr_curr)) != NULL) {
    /* strcpy is sufficient as ptr_curr->value is already checked for length */
    quoted_token = mstrdup(ptr_curr->value);
    if (!quoted_token) {
      error = 1;
      break;
    }
    if (dbi_conn_quote_string(conn, &quoted_token) == 0) {
      error = 1;
      break;
    }

    if (!strcmp(ptr_curr->name, ":JO:")) {
      strcpy(jo, quoted_token);
    }
    else if (!strcmp(ptr_curr->name, ":JF:")) {
      strcpy(jf, quoted_token);
    }
    else if (!strcmp(ptr_curr->name, ":J1:")) {
      strcpy(j1, quoted_token);
    }
    else if (!strcmp(ptr_curr->name, ":J2:")) {
      strcpy(j2, quoted_token);
    }
    free(quoted_token);
  }

  if (error) {
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    retval = 1;
    goto Finish;
  }

  /* assemble query string */
  strcpy(sql_command, "UPDATE t_periodical SET ");

  if (*jf) {
    sprintf(sql_command + strlen(sql_command), "periodical_name=%s, ", jf);
    syn_count++;
    sprintf(return_msg, "425:%s:%s\n", journal_val, jf);
  }

  if (*jo) {
    sprintf(sql_command + strlen(sql_command), "periodical_abbrev=%s, ", jo);
    syn_count++;
    sprintf(return_msg+strlen(return_msg), "425:%s:%s\n", journal_val, jo);
  }

  if (*j1) {
    sprintf(sql_command + strlen(sql_command), "periodical_custabbrev1=%s, ", j1);
    syn_count++;
    sprintf(return_msg+strlen(return_msg), "425:%s:%s\n", journal_val, j1);
  }

  if (*j2) {
    sprintf(sql_command + strlen(sql_command), "periodical_custabbrev2=%s, ", j2);
    syn_count++;
    sprintf(return_msg+strlen(return_msg), "425:%s:%s\n", journal_val, j2);
  }

  /* remove trailing comma */
  sprintf(sql_command  + strlen(sql_command) - 2, " WHERE %s=%s", journal_field, journal_val);

  LOG_PRINT(LOG_DEBUG, sql_command);

  dbires = dbi_conn_query(conn, sql_command);

  if (!dbires) {
    error = 1;
    goto Finish;
  }
    
  ptr_addresult->success += syn_count;

 Finish:
  if (ptr_addresult->success) {
    /* database was changed, update meta info */
    update_meta(conn, ptr_clrequest);
  }

  /* send back result message */
  if (!error) {
    send_status(ptr_clrequest->fd, 0, TERM_NO);
  }
  else {
    send_status(ptr_clrequest->fd, 260, TERM_NO);
  }

  tiwrite(ptr_clrequest->fd, return_msg, TERM_YES); 

  dbi_conn_close(conn);
  delete_all_lilimem(&sentinel);
  delete_all_liliform(&syn_sentinel);
  
  return retval;
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  is_journal(): checks whether journal exists

  int is_journal returns 0 = not found, 1 = journal exists, -1: error

  dbi_conn conn connection to database

  const char* field field to check

  const char* quoted_name periodical name (properly escaped and quoted)

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int is_journal(dbi_conn conn, const char* field, const char* quoted_name) {
  unsigned long long result;
  char sql_command[512];
  dbi_result dbires;

  snprintf(sql_command, 512, "SELECT periodical_id FROM t_periodical WHERE %s=%s", field, quoted_name);

  LOG_PRINT(LOG_DEBUG, sql_command);

  dbires = dbi_conn_query(conn, sql_command);

  if (!dbires) {
    LOG_PRINT(LOG_ERR, get_status_msg(234));
    return -1;
  }

  result = dbi_result_get_numrows(dbires);

  if (!result) {
    return 0;
  }
  else if (result == DBI_ROW_ERROR) {
    LOG_PRINT(LOG_ERR, get_status_msg(234));
    return -1;
  }
  else {
    return 1;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  getref(): implements the client command getref

  int getref returns 0 = no error, 1 = out of memory or db open error, 2 = subselect failed

  struct CLIENT_REQUEST* ptr_clrequest ptr to structure with client info

  struct bibinfo *ptr_biblio_info ptr to structure holding bibliography style

  int ref_format the output format 0 = simple screen rendering,
        1 = RIS format, 2 = DocBook format, 3 = BibTeX format,
        4 = HTML 10 = RISX 11 = XHTML

  int n_privatelist if set to 1, limit search to user's private list
                    if set to 0, search all

  char* pdfroot root path used for PDF or Postscript offprint paths

  char* cgi_url URL of a CSS file used to customize HTML output

  struct ADDRESULT* ptr_addresult ptr to result counter structure

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int getref(struct CLIENT_REQUEST* ptr_clrequest, struct bibinfo *ptr_biblio_info, int ref_format, int n_privatelist, char *pdfroot, char *cgi_url, struct ADDRESULT* ptr_addresult) {
  dbi_conn conn;
  dbi_result dbires;
  dbi_driver driver;
  char **inargv; /* tokens of the client request */
  int inargc; /* number of tokens of the client request */
  int inargcmax; /* maximum number of tokens of the client request */
  int len;
  int counter;
  int n_isalpha;
  int n_havesome;
  int numbyte; /* number of bytes written */
  int author_type; /* signals type of author query, if any */
  int render_res;
  int cs_status;
  int retval = 0;
  unsigned long long n_id;
  unsigned long long nref_counter = 0;
  size_t buffer_len; /* these are the allocated lengths of the buffers */
  size_t sql_command_len;
  size_t sql_command1_len;
  size_t stringbuf_len;
  size_t result_len;
  size_t n_bufsize;
  char *sql_command; /* these are ptrs to the buffers and temporary ptrs */
  char *new_sql_command;
  char *sql_command1;
  char *new_sql_command1;
  char *buffer; 
  char *new_buffer;
  char *stringbuf;
  char *new_stringbuf;
  char *token;
  char *newtoken;
  char *quoted_token;
  char *eostring;
  char *query_string;
  const char* db_encoding;
  char operator[5];
  char the_id[32];
  char bitsandpieces[256];
  struct SQLTOKEN sqltoken;
  struct lilimem sentinel;
  struct renderinfo rendinfo;
  iconv_t conv_descriptor;

  sentinel.ptr_mem = NULL;
  sentinel.ptr_next = NULL;
  sentinel.varname[0] = '\0';

  /* fill in invariant elements of structure */
  rendinfo.ptr_biblio_info = ptr_biblio_info;
  rendinfo.ref_format = ref_format;
  rendinfo.nuse_citestyle = 0;
  rendinfo.database = ptr_clrequest->current_db;
  rendinfo.dbname = NULL;
  rendinfo.username = ptr_clrequest->username;
  rendinfo.pdfroot = pdfroot;
  rendinfo.cgi_url = cgi_url;
  rendinfo.ptr_clrequest = ptr_clrequest;

  /* get buffer to analyze the request */
  inargc = 0;
  inargcmax = 10;
  inargv = malloc((size_t)inargcmax*sizeof(char*));
  if (inargv == NULL || insert_lilimem(&sentinel, (void**)&inargv, NULL)) {
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 1;
  }

  /* get buffer to assemble the SQL queries */
  sql_command_len = 4096; 
  sql_command = malloc(sql_command_len); 

  if (sql_command == NULL || insert_lilimem(&sentinel, (void**)&sql_command, NULL)) {
    delete_all_lilimem(&sentinel);
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 1;
  }
  sql_command[0] = '\0'; /* start with an empty string */

  /* get another buffer to assemble the SQL queries */
  sql_command1_len = 4096;
  sql_command1 = malloc(sql_command1_len); 

  if (sql_command1 == NULL || insert_lilimem(&sentinel, (void**)&sql_command1, NULL)) {
    delete_all_lilimem(&sentinel);
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 1;
  }
  sql_command1[0] = '\0'; /* start with an empty string */

  /* get a buffer to hold tokens */
  buffer_len = 4096;
  buffer = malloc(buffer_len); 

  if (buffer == NULL || insert_lilimem(&sentinel, (void**)&buffer, NULL)) {
    delete_all_lilimem(&sentinel);
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 1;
  }
  buffer[0] = '\0'; /* start with an empty string */

  /* get a buffer to hold tokens */
  stringbuf_len = 4096;
  stringbuf = malloc(stringbuf_len); 

  if (stringbuf == NULL || insert_lilimem(&sentinel, (void**)&stringbuf, NULL)) {
    delete_all_lilimem(&sentinel);
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 1;
  }
  buffer[0] = '\0'; /* start with an empty string */

  /* get length of the input string */
  n_bufsize = atoi(ptr_clrequest->argument);

  /* try to allocate the amount the client requested */
  query_string = malloc(n_bufsize);
  if (query_string == NULL || insert_lilimem(&sentinel, (void**)&query_string, NULL)) {
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 1;
  }

  /* send acknowledgement to client */
  send_status(ptr_clrequest->fd, 0, TERM_NO);

  if ((cs_status = read_status(ptr_clrequest->fd)) != 0) {
    LOG_PRINT(LOG_INFO, get_status_msg(112));
    delete_all_lilimem(&sentinel);
    return 1;
  }

  /* read id list from client */
  numbyte = tread(ptr_clrequest->fd, query_string, n_bufsize);
  if (numbyte == -1) {
    LOG_PRINT(LOG_INFO, get_status_msg(109));
    delete_all_lilimem(&sentinel);
    return 1;
  }

/*   printf("query: %s<<\n", query_string); */
  eostring = query_string + strlen(query_string);

  /* connect to the database */
  if ((conn = connect_to_db(ptr_clrequest, NULL, 0)) == NULL) {
    send_status(ptr_clrequest->fd, 204, TERM_NO);
    LOG_PRINT(LOG_WARNING, get_status_msg(204));
    delete_all_lilimem(&sentinel);
    return 1;
  }

  driver = dbi_conn_get_driver(conn);

  /* get the database encoding */
  db_encoding = dbi_conn_get_encoding(conn);

  /* if we need to convert, create a conversion descriptor for iconv() */
  if (db_encoding && *(ptr_biblio_info->encoding) && strcmp(db_encoding, ptr_biblio_info->encoding)) {
    conv_descriptor = iconv_open(!strcmp(ptr_biblio_info->encoding, "US-ASCII") ? "ASCII" : ptr_biblio_info->encoding, !strcmp(db_encoding, "US-ASCII") ? "ASCII" : db_encoding);
    if (conv_descriptor == (iconv_t)(-1)) {
      LOG_PRINT(LOG_WARNING, "cannot set conversion descriptor (database/output):");
      LOG_PRINT(LOG_DEBUG, db_encoding);
      LOG_PRINT(LOG_DEBUG, ptr_biblio_info->encoding);
      send_status(ptr_clrequest->fd, 701, TERM_NO);
      LOG_PRINT(LOG_WARNING, get_status_msg(701));
      delete_all_lilimem(&sentinel);
      return 1;
    }
    else {
      LOG_PRINT(LOG_DEBUG, "database encoding is:");
      LOG_PRINT(LOG_DEBUG, db_encoding);
    }
  }
  else {
    conv_descriptor = NULL;
    LOG_PRINT(LOG_DEBUG, "no character encoding conversion required");
  }

  LOG_PRINT(LOG_DEBUG, "output encoding is:");
  if (*(ptr_biblio_info->encoding)) {
    LOG_PRINT(LOG_DEBUG, ptr_biblio_info->encoding);
  }
  else {
    LOG_PRINT(LOG_DEBUG, db_encoding);
  }


  token = query_string;
  newtoken = token;

  /* general strategy: we extract the tokens one after the other and insert
     brackets, logical operators, and those requests which map to simple
     t_refdb queries directly into the future search string. For those 
     queries that require a relational lookup a surrogate subselect query
     is formulated, as mySQL does not (yet) support true subselects.
     The results of these subselects are lists of refdb_id values which
     are translated into query fragments and get inserted into the future
     search string. The final search string is sent to the db server
     and provides us with the final list of the t_refdb fields we're
     interested in. This list is finally used to extract the additional
     data from the t_author, t_keyword, and t_user fields that we need to
     send back the result. */

  /* loop as long as we find more tokens */
  while (newtoken != NULL) {
    token = sql_tokenize(newtoken, &sqltoken);
    newtoken = sqltoken.next_token;
/*      printf("%s\n", token);  */

    if (token != NULL) {
      /* extract the token and save to a temporary string */
      if (sqltoken.length > buffer_len) {
	new_buffer = (char*)realloc(buffer, (size_t)sqltoken.length);
	if (new_buffer == NULL) { /* out of memory */
	  retval = 1;
	  break;
	}
	else {
	  buffer = new_buffer;
	  buffer_len = sqltoken.length;
	}
      }
      strncpy(buffer, token, sqltoken.length);
      buffer[sqltoken.length] ='\0';
      if (sqltoken.type != 4) { /* we recycle these tokens */
	if ((new_sql_command1 = mstrcat(sql_command1, buffer, &sql_command1_len, 0)) == NULL) {
	  retval = 1;
	  break;
	}
	else {
	  sql_command1 = new_sql_command1;
	}
      }
      else {
	/*----------------------------------------------------------------*/
	/* author */
	if (strncmp(buffer, ":AU:", 4) == 0 || strncmp(buffer, ":A1:", 4) == 0) {
	  author_type = 1;
	}
	else if (strncmp(buffer, ":A2:", 4) == 0 || strncmp(buffer, ":ED:", 4) == 0) {
	  author_type = 2;
	}
	else if (strncmp(buffer, ":A3:", 4) == 0) {
	  author_type = 3;
	}
	else if (strncmp(buffer, ":AX:", 4) == 0) {
	  author_type = 4;
	}
	else {
	  author_type = 0;
	}

	if (author_type != 0) { /* author query */
	  strncpy(operator, &token[4], sqltoken.length-4); /* save the comparison operator */
	  operator[sqltoken.length-4] = '\0'; /* terminate string */

	  /* assemble a surrogate subselect query */
	  token = sql_tokenize(newtoken, &sqltoken);
	  newtoken = sqltoken.next_token;
	  if (token != NULL) {
	    char author_type_start[64];
	    char author_type_end[4];

	    strncpy(buffer, token, sqltoken.length);
	    buffer[sqltoken.length] ='\0'; /* terminate string */
	    len = 0;
	    token = nstrtok(buffer, &len, " ");
	    if (token != NULL) {

	      /* prepare bits and pieces for AND/OR queries*/
	      if (author_type == 4) { /* any type of author */
		*author_type_start = '\0';
		strcpy(author_type_end, ") ");
	      }
	      else { /* particular type of author */
		sprintf(author_type_start, "(t_xauthor.xauthor_type=%s AND ", get_author_type_string(driver, author_type));
		strcpy(author_type_end, ")) ");
	      }

	      /* plain query with one author */
	      if (*token != '&' && *token != '|') {
		if (author_type == 4) { /* any type of author */
		  sprintf(sql_command, "SELECT DISTINCT t_xauthor.refdb_id FROM t_xauthor, t_author WHERE t_author.author_id=t_xauthor.author_id AND t_author.author_name %s ", (*operator == '=' || (*operator == '!' && *(operator+1) == '=')) ? "=" : my_dbi_conn_get_cap(conn, "rlike"));
		}
		else { /* particular type of author */
		  sprintf(sql_command, "SELECT DISTINCT t_xauthor.refdb_id FROM t_xauthor, t_author WHERE t_author.author_id=t_xauthor.author_id AND t_xauthor.xauthor_type=%s AND t_author.author_name %s ", get_author_type_string(driver, author_type), (*operator == '=' || (*operator == '!' && *(operator+1) == '=')) ? "=" : my_dbi_conn_get_cap(conn, "rlike"));
		}
		quoted_token = mstrdup(token);
		if (!quoted_token) {
		  retval = 1;
		  break;
		}
		if (dbi_conn_quote_string(conn, &quoted_token) == 0) {
		  retval = 1;
		  break;
		}

		if ((new_sql_command = mstrcat(sql_command, quoted_token, &sql_command_len, 0)) == NULL) {
		  retval = 1;
		  free(quoted_token);
		  break;
		}
		else {
		  sql_command = new_sql_command;
		}
		free(quoted_token);
	      }

	      /* OR query with several authors */
	      else if (*token == '|') {
		sprintf(sql_command, "SELECT DISTINCT t_xauthor.refdb_id FROM t_xauthor, t_author WHERE %s (", author_type_start);

		while (token != NULL) {
		  token = nstrtok(token+len, &len, " ");
		  if (token != NULL) {
		    if (*operator == '=' || (*operator == '!' && *(operator+1) == '=')) {
		      strcpy(bitsandpieces, "t_author.author_name = ");
		    }
		    else {
		      sprintf(bitsandpieces, "t_author.author_name %s ", my_dbi_conn_get_cap(conn, "rlike"));
		    }

		    if ((new_sql_command = mstrcat(sql_command, bitsandpieces, &sql_command_len, 0)) == NULL) {
		      retval = 1;
		      break;
		    }
		    else {
		      sql_command = new_sql_command;
		    }
		    if ((new_stringbuf = mstrncpy(stringbuf, token, len, &stringbuf_len)) == NULL) {
		      retval = 1;
		      break;
		    }
		    else {
		      stringbuf = new_stringbuf;
		    }
		    stringbuf[len] = '\0';
		    quoted_token = mstrdup(stringbuf);
		    if (!quoted_token) {
		      retval = 1;
		      break;
		    }
		    if (dbi_conn_quote_string(conn, &quoted_token) == 0) {
		      retval = 1;
		      break;
		    }
		    if ((new_sql_command = mstrcat(sql_command, quoted_token, &sql_command_len, 0)) == NULL) {
		      retval = 1;
		      free(quoted_token);
		      break;
		    }
		    else {
		      sql_command = new_sql_command;
		    }
		    free(quoted_token);
		    if ((new_sql_command = mstrcat(sql_command, " OR ", &sql_command_len, 0)) == NULL) {
		      retval = 1;
		      break;
		    }
		    else {
		      sql_command = new_sql_command;
		    }
		  }
		  else {
		    if ((new_sql_command = mstrcat(sql_command, author_type_end, &sql_command_len, 4)) == NULL) {
		      retval = 1;
		      break;
		    }
		    else {
		      sql_command = new_sql_command;
		    }
		  }
		} /* end while */
		if (retval) {
		  break;
		}

		if ((new_sql_command = mstrcat(sql_command, "AND t_author.author_id=t_xauthor.author_id", &sql_command_len, 0)) == NULL) {
		  retval = 1;
		  break;
		}
		else {
		  sql_command = new_sql_command;
		}
	      }

	      /* AND query with several authors */
	      else if (*token == '&') {
		counter = 0;
		sprintf(sql_command, "SELECT DISTINCT t_xauthor.refdb_id, COUNT(*) AS count FROM t_xauthor, t_author WHERE %s (", author_type_start);

		while (token != NULL) {
		  token = nstrtok(token+len, &len, " ");
		  if (token != NULL) {
		    if (*operator == '=' || (*operator == '!' && *(operator+1) == '=')) {
		      strcpy(bitsandpieces, "t_author.author_name = ");
		    }
		    else {
		      sprintf(bitsandpieces, "t_author.author_name %s ", my_dbi_conn_get_cap(conn, "rlike"));
		    }

		    if ((new_sql_command = mstrcat(sql_command, bitsandpieces, &sql_command_len, 0)) == NULL) {
		      retval = 1;
		      break;
		    }
		    else {
		      sql_command = new_sql_command;
		    }
		    if ((new_stringbuf = mstrncpy(stringbuf, token, len, &stringbuf_len)) == NULL) {
		      retval = 1;
		      break;
		    }
		    else {
		      stringbuf = new_stringbuf;
		    }
		    stringbuf[len] = '\0';
		    quoted_token = mstrdup(stringbuf);
		    if (!quoted_token) {
		      retval = 1;
		      break;
		    }
		    if (dbi_conn_quote_string(conn, &quoted_token) == 0) {
		      retval = 1;
		      break;
		    }
		    if ((new_sql_command = mstrcat(sql_command, quoted_token, &sql_command_len, 0)) == NULL) {
		      retval = 1;
		      free(quoted_token);
		      break;
		    }
		    else {
		      sql_command = new_sql_command;
		    }
		    free(quoted_token);
		    if ((new_sql_command = mstrcat(sql_command, " OR ", &sql_command_len, 0)) == NULL) {
		      retval = 1;
		      break;
		    }
		    else {
		      sql_command = new_sql_command;
		    }
		    counter++;
		  }
		  else {
		    if ((new_sql_command = mstrcat(sql_command, author_type_end, &sql_command_len, 4)) == NULL) {
		      retval = 1;
		      break;
		    }
		    else {
		      sql_command = new_sql_command;
		    }
		  }
		}
		if ((new_sql_command = mstrcat(sql_command, "AND t_author.author_id=t_xauthor.author_id GROUP BY t_xauthor.refdb_id HAVING count=", &sql_command_len, 0)) == NULL) {
		  retval = 1;
		  break;
		}
		else {
		  sql_command = new_sql_command;
		}
		sprintf(buffer, "%d", counter);
		if ((new_sql_command = mstrcat(sql_command, buffer, &sql_command_len, 0)) == NULL) {
		  retval = 1;
		  break;
		}
		else {
		  sql_command = new_sql_command;
		}
	      }
	    }
	  }

	  LOG_PRINT(LOG_DEBUG, sql_command);

	  /* dbires will contain all refdb_ids of those entries that 
	     match the author requirements in the pseudo-subselect */
	  dbires = dbi_conn_query(conn, sql_command);
	  if (!dbires) {
	    retval = 2;
	    break;
	  }

	  if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_id IN (", &sql_command1_len, 0)) == NULL) {
	    retval = 1;
	    break;
	  }
	  else {
	    sql_command1 = new_sql_command1;
	  }

	  n_havesome = 0;
	  while (dbi_result_next_row(dbires) != 0) {
	    n_havesome = 1;
	    n_id = my_dbi_result_get_idval_idx(dbires, 1); /* 1-base index */
	    sprintf(the_id, ULLSPEC, (unsigned long long)n_id);
	    if ((new_sql_command1 = mstrcat(sql_command1, the_id, &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      dbi_result_free(dbires);
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	    if ((new_sql_command1 = mstrcat(sql_command1, ",", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      dbi_result_free(dbires);
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }

	  if (!n_havesome) { /* use a dummy value that does not exist in the database */
	    if ((new_sql_command1 = mstrcat(sql_command1, "-1,", &sql_command1_len, 0)) == NULL) { /* the comma will be removed further down */
	      retval = 1;
	      dbi_result_free(dbires);
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }	    
	  dbi_result_free(dbires);
	  if (retval) {
	    break;
	  }
	  if ((new_sql_command1 = mstrcat(sql_command1, ")", &sql_command1_len, 1)) == NULL) {
	    retval = 1;
	    break;
	  }
	  else {
	    sql_command1 = new_sql_command1;
	  }
	}

	/*----------------------------------------------------------------*/
	/* keyword */
	else if (strncmp(token, ":KW:", 4) == 0) { /* keyword query */
	  strncpy(operator, &token[4], sqltoken.length-4); /* save the comparison operator */
	  operator[sqltoken.length-4] = '\0'; /* terminate string */

	  /* assemble a surrogate subselect query */
	  token = sql_tokenize(newtoken, &sqltoken);
	  newtoken = sqltoken.next_token;
	  if (token != NULL) {
	    strncpy(buffer, token, sqltoken.length);
	    buffer[sqltoken.length] ='\0'; /* terminate string */
	    len = 0;
	    token = nstrtok(buffer, &len, " ");
	    if (token != NULL) {
	      if (*token != '&' && *token != '|') {
		/* simple keyword query */
		sprintf(sql_command, "SELECT DISTINCT t_xkeyword.xref_id FROM t_xkeyword, t_keyword WHERE t_xkeyword.xkeyword_type=\'REFERENCE\' AND t_keyword.keyword_id=t_xkeyword.keyword_id AND t_keyword.keyword_name %s ", (*operator == '=' || (*operator == '!' && *(operator+1) == '=')) ? "=" : my_dbi_conn_get_cap(conn, "rlike"));
		quoted_token = mstrdup(token);
		if (!quoted_token) {
		  retval = 1;
		  break;
		}
		if (dbi_conn_quote_string(conn, &quoted_token) == 0) {
		  retval = 1;
		  break;
		}
		if ((new_sql_command = mstrcat(sql_command, quoted_token, &sql_command_len, 0)) == NULL) {
		  retval = 1;
		  free(quoted_token);
		  break;
		}
		else {
		  sql_command = new_sql_command;
		}
		free(quoted_token);
	      }
	      else if (*token == '|') {
		strcpy(sql_command, "SELECT DISTINCT t_xkeyword.xref_id FROM t_xkeyword, t_keyword WHERE t_xkeyword.xkeyword_type=\'REFERENCE\' AND (");
		while (token != NULL) {
		  token = nstrtok(token+len, &len, " ");
		  if (token != NULL) {
		    if (*operator == '=' || (*operator == '!' && *(operator+1) == '=')) {
		      strcpy(bitsandpieces, "t_keyword.keyword_name = ");
		    }
		    else {
		      sprintf(bitsandpieces, "t_keyword.keyword_name %s ", my_dbi_conn_get_cap(conn, "rlike"));
		    }

		    if ((new_sql_command = mstrcat(sql_command, bitsandpieces, &sql_command_len, 0)) == NULL) {
		      retval = 1;
		      break;
		    }
		    else {
		      sql_command = new_sql_command;
		    }
		    if ((new_stringbuf = mstrncpy(stringbuf, token, len, &stringbuf_len)) == NULL) {
		      retval = 1;
		      break;
		    }
		    else {
		      stringbuf = new_stringbuf;
		    }
		    stringbuf[len] = '\0';
		    quoted_token = mstrdup(stringbuf);
		    if (!quoted_token) {
		      retval = 1;
		      break;
		    }
		    if (dbi_conn_quote_string(conn, &quoted_token) == 0) {
		      retval = 1;
		      break;
		    }
		    if ((new_sql_command = mstrcat(sql_command, quoted_token, &sql_command_len, 0)) == NULL) {
		      retval = 1;
		      free(quoted_token);
		      break;
		    }
		    else {
		      sql_command = new_sql_command;
		    }
		    free(quoted_token);

		    if ((new_sql_command = mstrcat(sql_command, " OR ", &sql_command_len, 0)) == NULL) {
		      retval = 1;
		      break;
		    }
		    else {
		      sql_command = new_sql_command;
		    }
		  }
		  else {
		    if ((new_sql_command = mstrcat(sql_command, ") ", &sql_command_len, 4)) == NULL) {
		      retval = 1;
		      break;
		    }
		    else {
		      sql_command = new_sql_command;
		    }
		  }
		}
		if ((new_sql_command = mstrcat(sql_command, "AND t_keyword.keyword_id=t_xkeyword.keyword_id", &sql_command_len, 0)) == NULL) {
		  retval = 1;
		  break;
		}
		else {
		  sql_command = new_sql_command;
		}
	      }
	      else if (*token == '&') {
		counter = 0;
		strcpy(sql_command, "SELECT DISTINCT t_xkeyword.xref_id, COUNT(*) AS count FROM t_xkeyword, t_keyword WHERE t_xkeyword.xkeyword_type=\'REFERENCE\' AND (");
		while (token != NULL) {
		  token = nstrtok(token+len, &len, " ");
		  if (token != NULL) {
		    if (*operator == '=' || (*operator == '!' && *(operator+1) == '=')) {
		      strcpy(bitsandpieces, "t_keyword.keyword_name = ");
		    }
		    else {
		      sprintf(bitsandpieces, "t_keyword.keyword_name %s ", my_dbi_conn_get_cap(conn, "rlike"));
		    }

		    if ((new_sql_command = mstrcat(sql_command, bitsandpieces, &sql_command_len, 0)) == NULL) {
		      retval = 1;
		      break;
		    }
		    else {
		      sql_command = new_sql_command;
		    }
		    if ((new_stringbuf = mstrncpy(stringbuf, token, len, &stringbuf_len)) == NULL) {
		      retval = 1;
		      break;
		    }
		    else {
		      stringbuf = new_stringbuf;
		    }
		    stringbuf[len] = '\0';
		    quoted_token = mstrdup(stringbuf);
		    if (!quoted_token) {
		      retval = 1;
		      break;
		    }
		    if (dbi_conn_quote_string(conn, &quoted_token) == 0) {
		      retval = 1;
		      break;
		    }
		    if ((new_sql_command = mstrcat(sql_command, quoted_token, &sql_command_len, 0)) == NULL) {
		      retval = 1;
		      free(quoted_token);
		      break;
		    }
		    else {
		      sql_command = new_sql_command;
		    }
		    free(quoted_token);

		    if ((new_sql_command = mstrcat(sql_command, " OR ", &sql_command_len, 0)) == NULL) {
		      retval = 1;
		      break;
		    }
		    else {
		      sql_command = new_sql_command;
		    }
		    counter++;
		  }
		  else {
		    if ((new_sql_command = mstrcat(sql_command, ") ", &sql_command_len, 4)) == NULL) {
		      retval = 1;
		      break;
		    }
		    else {
		      sql_command = new_sql_command;
		    }
		  }
		}
		if ((new_sql_command = mstrcat(sql_command, "AND t_keyword.keyword_id=t_xkeyword.keyword_id GROUP BY t_xkeyword.xref_id HAVING count=", &sql_command_len, 0)) == NULL) {
		  retval = 1;
		  break;
		}
		else {
		  sql_command = new_sql_command;
		}
		sprintf(buffer, "%d", counter);
		if ((new_sql_command = mstrcat(sql_command, buffer, &sql_command_len, 0)) == NULL) {
		  retval = 1;
		  break;
		}
		else {
		  sql_command = new_sql_command;
		}
	      }
	    }
	  }

	  LOG_PRINT(LOG_DEBUG, sql_command);

	  /* dbires will contain all refdb_ids that match the above
	     keyword requirements in the pseudo-subselect */
	  dbires = dbi_conn_query(conn, sql_command);
	  if (!dbires) {
	    retval = 2;
	    break;
	  }

	  if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_id IN (", &sql_command1_len, 0)) == NULL) {
	    retval = 1;
	    break;
	  }
	  else {
	    sql_command1 = new_sql_command1;
	  }

	  n_havesome = 0;
	  while (dbi_result_next_row(dbires) != 0) {
	    n_havesome = 1;
	    n_id = my_dbi_result_get_idval_idx(dbires, 1);
	    sprintf(the_id, ULLSPEC, (unsigned long long)n_id);
	    if ((new_sql_command1 = mstrcat(sql_command1, the_id, &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      dbi_result_free(dbires);
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	    if ((new_sql_command1 = mstrcat(sql_command1, ",", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      dbi_result_free(dbires);
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  if (!n_havesome) { /* use a dummy value that does not exist in the database */
	    if ((new_sql_command1 = mstrcat(sql_command1, "-1,", &sql_command1_len, 0)) == NULL) { /* the comma will be removed further down */
	      retval = 1;
	      dbi_result_free(dbires);
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }	    

	  dbi_result_free(dbires);
	  if (retval) {
	    break;
	  }
	  if ((new_sql_command1 = mstrcat(sql_command1, ")", &sql_command1_len, 1)) == NULL) {
	    retval = 1;
	    break;
	  }
	  else {
	    sql_command1 = new_sql_command1;
	  }
	}
	/*----------------------------------------------------------------*/
	/* journal */
	else if (strncmp(token, ":JO:", 4) == 0
		 || strncmp(token, ":JF:", 4) == 0
		 || strncmp(token, ":J1:", 4) == 0
		 || strncmp(token, ":J2:", 4) == 0
		 || strncmp(token, ":JX:", 4) == 0) { /* periodical query */
	  char my_token[5]; /* backup of token */
	  char my_bool[4];
	  strncpy(my_token, token, 4);
	  my_token[4] = '\0';
	  strncpy(operator, &token[4], sqltoken.length-4); /* save the comparison operator */
	  operator[sqltoken.length-4] = '\0'; /* terminate string */

	  /* translate operator to regexp operator */
	  if (*operator == '=') {
	    strcpy(bitsandpieces, " = ");
	    strcpy(my_bool, "OR");
	  }
	  else if (*operator == '~') {
	    sprintf(bitsandpieces, " %s ", my_dbi_conn_get_cap(conn, "rlike"));
	    strcpy(my_bool, "OR");
	  }
	  else if (*(operator+1) == '=') { 
	    strcpy(bitsandpieces, " != ");
	    strcpy(my_bool, "AND");
	  }
	  else { /* treat all other operators as non-equal regexp*/
	    sprintf(bitsandpieces, " %s ", my_dbi_conn_get_cap(conn, "not_rlike"));
	    strcpy(my_bool, "AND");
	  }
	    
	  token = sql_tokenize(newtoken, &sqltoken);
	  newtoken = sqltoken.next_token;
	  if (token != NULL) {
	    if ((new_buffer = mstrncpy(buffer, token, sqltoken.length, &buffer_len)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      buffer = new_buffer;
	    }
	    buffer[sqltoken.length] ='\0';
	    quoted_token = mstrdup(buffer);
	    if (!quoted_token) {
	      retval = 1;
	      break;
	    }
	    if (dbi_conn_quote_string(conn, &quoted_token) == 0) {
	      retval = 1;
	      break;
	    }
	  }
	  else {
	    quoted_token = mstrdup("");
	  }

	  /* assemble a surrogate subselect query */
	  if (strncmp(my_token, ":JX:", 4) == 0) {
	    char jxbuffer[2048];
	    snprintf(jxbuffer, 2048, "SELECT t_refdb.refdb_periodical_id FROM t_refdb, t_periodical WHERE t_periodical.periodical_id=t_refdb.refdb_periodical_id AND (t_periodical.periodical_name%s%s %s t_periodical.periodical_abbrev%s%s %s t_periodical.periodical_custabbrev1%s%s %s t_periodical.periodical_custabbrev2%s%s)", bitsandpieces, quoted_token, my_bool, bitsandpieces, quoted_token, my_bool, bitsandpieces, quoted_token, my_bool, bitsandpieces, quoted_token);
	    free(quoted_token);
	    *sql_command = '\0';
	    if ((new_sql_command = mstrcat(sql_command, jxbuffer, &sql_command_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command = new_sql_command;
	    }
	  }
	  else {
	    if (strncmp(my_token, ":JF:", 4) == 0) {
	      strcpy(sql_command, "SELECT t_refdb.refdb_periodical_id FROM t_refdb, t_periodical WHERE t_periodical.periodical_id=t_refdb.refdb_periodical_id AND t_periodical.periodical_name");
	    }
	    else if (strncmp(my_token, ":JO:", 4) == 0) {
	      strcpy(sql_command, "SELECT t_refdb.refdb_periodical_id FROM t_refdb, t_periodical WHERE t_periodical.periodical_id=t_refdb.refdb_periodical_id AND t_periodical.periodical_abbrev");
	    }
	    else if (strncmp(my_token, ":J1:", 4) == 0) {
	      strcpy(sql_command, "SELECT t_refdb.refdb_periodical_id FROM t_refdb, t_periodical WHERE t_periodical.periodical_id=t_refdb.refdb_periodical_id AND t_periodical.periodical_custabbrev1");
	    }
	    else if (strncmp(my_token, ":J2:", 4) == 0){
	      strcpy(sql_command, "SELECT t_refdb.refdb_periodical_id FROM t_refdb, t_periodical WHERE t_periodical.periodical_id=t_refdb.refdb_periodical_id AND t_periodical.periodical_custabbrev2");
	    }

	    if ((new_sql_command = mstrcat(sql_command, bitsandpieces, &sql_command_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command = new_sql_command;
	    }
	    
	    if (quoted_token != NULL) {
	      if ((new_sql_command = mstrcat(sql_command, quoted_token, &sql_command_len, 0)) == NULL) {
		retval = 1;
		free(quoted_token);
		break;
	      }
	      else {
		sql_command = new_sql_command;
	      }
	      free(quoted_token);
	    }
	  } /* end if JX */

	  LOG_PRINT(LOG_DEBUG, sql_command);
	  dbires = dbi_conn_query(conn, sql_command);
	  if (!dbires) {
	    retval = 2;
	    break;
	  }

	  if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_periodical_id IN (", &sql_command1_len, 0)) == NULL) {
	    retval = 1;
	    break;
	  }
	  else {
	    sql_command1 = new_sql_command1;
	  }
/* 	  if ((new_sql_command1 = mstrcat(sql_command1, operator, &sql_command1_len, 0)) == NULL) { */
/* 	    error = 1; */
/* 	    break; */
/* 	  } */
/* 	  else { */
/* 	    sql_command1 = new_sql_command1; */
/* 	  } */

	  if (dbi_result_get_numrows(dbires)) {
	    while (dbi_result_next_row(dbires) != 0) {
	      n_id = my_dbi_result_get_idval_idx(dbires, 1);
	      sprintf(the_id, ULLSPEC, (unsigned long long)n_id);
	      if ((new_sql_command1 = mstrcat(sql_command1, the_id, &sql_command1_len, 0)) == NULL) {
		retval = 1;
		dbi_result_free(dbires);
		break;
	      }
	      else {
		sql_command1 = new_sql_command1;
	      }
	      if ((new_sql_command1 = mstrcat(sql_command1, ",", &sql_command1_len, 0)) == NULL) {
		retval = 1;
		dbi_result_free(dbires);
		break;
	      }
	      else {
		sql_command1 = new_sql_command1;
	      }
	    }
	    /* remove trailing comma */
	    sql_command1[strlen(sql_command1)-1] = '\0';
	  }
	  else { /* use a dummy value that does not exist in the database */
	    if ((new_sql_command1 = mstrcat(sql_command1, "-1", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      dbi_result_free(dbires);
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }	    
	  
	  if ((new_sql_command1 = mstrcat(sql_command1, ")", &sql_command1_len, 0)) == NULL) {
	    retval = 1;
	    dbi_result_free(dbires);
	    break;
	  }
	  else {
	    sql_command1 = new_sql_command1;
	  }
	  dbi_result_free(dbires);
	}

	/*----------------------------------------------------------------*/
	/* search in all titles */
	else if (strncmp(token, ":TA:", 4) == 0 /* deprecated */
		 || strncmp(token, ":TX:", 4) == 0) { /* all titles query */
	  char table[3][27] = {"t_refdb.refdb_title",
			    "t_refdb.refdb_booktitle",
			    "t_refdb.refdb_title_series"};
	  char the_op[6];
	  int i;
	  int n_isfirst = 1;

	  /* todo: fixme */
	  strncpy(operator, &token[4], sqltoken.length-4); /* save the comparison operator */
	  operator[sqltoken.length-4] = '\0'; /* terminate string */

	  if (*operator == '=') {
	    strcpy(bitsandpieces, " = ");
	    strcpy(the_op, " OR ");
	  }
	  else if (*operator == '~') {
	    sprintf(bitsandpieces, " %s ", my_dbi_conn_get_cap(conn, "rlike"));
	    strcpy(the_op, " OR ");
	  }
	  else if (*operator == '!' && *(operator+1) == '=') {
	    sprintf(bitsandpieces, " != ");
	    strcpy(the_op, " AND ");
	  }
	  else { /* treat all other operators as non-equal */
	    sprintf(bitsandpieces, " %s ", my_dbi_conn_get_cap(conn, "not_rlike"));
	    strcpy(the_op, " AND ");
	  }
	  
	  token = sql_tokenize(newtoken, &sqltoken);
	  newtoken = sqltoken.next_token;
	  if (token != NULL) {
/*  	    printf("%s:%d\n", token, error); */
/*  	    printf("%s\n", newtoken); */
	    if ((new_buffer = mstrncpy(buffer, token, sqltoken.length, &buffer_len)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      buffer = new_buffer;
	    }
	    buffer[sqltoken.length] ='\0';
	    quoted_token = mstrdup(buffer);
	    if (!quoted_token) {
	      retval = 1;
	      break;
	    }
	    if (dbi_conn_quote_string(conn, &quoted_token) == 0) {
	      retval = 1;
	      break;
	    }

	    for (i = 0; i<3; i++) {
	      if (!n_isfirst) {
		if ((new_sql_command1 = mstrcat(sql_command1, the_op, &sql_command1_len, 0)) == NULL) {
		  retval = 1;
		  break;
		}
		else {
		  sql_command1 = new_sql_command1;
		}
	      }
	      else {
		n_isfirst--;
		if ((new_sql_command1 = mstrcat(sql_command1, "(", &sql_command1_len, 0)) == NULL) {
		  retval = 1;
		  break;
		}
		else {
		  sql_command1 = new_sql_command1;
		}
	      }

	      if ((new_sql_command1 = mstrcat(sql_command1, table[i], &sql_command1_len, 0)) == NULL) {
		retval = 1;
		break;
	      }
	      else {
		sql_command1 = new_sql_command1;
	      }
	  
	      if ((new_sql_command1 = mstrcat(sql_command1, bitsandpieces, &sql_command1_len, 0)) == NULL) {
		retval = 1;
		break;
	      }
	      else {
		sql_command1 = new_sql_command1;
	      }
	      
	      if ((new_sql_command1 = mstrcat(sql_command1, quoted_token, &sql_command1_len, 0)) == NULL) {
		retval = 1;
		free(quoted_token);
		break;
	      }
	      else {
		sql_command1 = new_sql_command1;
	      }
	    } /* end for */

	    if ((new_sql_command1 = mstrcat(sql_command1, ")", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }

	    free(quoted_token);
	  }
	}

	/*----------------------------------------------------------------*/
	/* extended notes - note ID */
	else if (strncmp(token, ":NID:", 5) == 0) { /* extended note ID query */
	  strncpy(operator, &token[5], sqltoken.length-5); /* save the comparison operator */
	  operator[sqltoken.length-5] = '\0'; /* terminate string */

	  /* assemble a surrogate subselect query */
	  quoted_token = mstrdup(ptr_clrequest->username);
	  if (!quoted_token) {
	    retval = 1;
	    break;
	  }
	  if (dbi_conn_quote_string(conn, &quoted_token) == 0) {
	    retval = 1;
	    break;
	  }

	  sprintf(sql_command, "SELECT t_refdb.refdb_id FROM t_refdb, t_note, t_xnote, t_user WHERE t_note.note_user_id=t_user.user_id AND t_user.user_name=%s AND t_refdb.refdb_id=t_xnote.xref_id AND t_xnote.xnote_type=\'REFERENCE\' AND t_note.note_id=t_xnote.note_id AND t_note.note_id", quoted_token);

	  free(quoted_token);

	  /* append operator */
	  if (*operator == '~') {
	    strcpy(bitsandpieces, "=");
	  }
	  else if (*operator == '!' && *(operator+1) == '~') {
	    strcpy(bitsandpieces, "!=");
	  }
	  else {
	    strcpy(bitsandpieces, operator);
	  }

	  if ((new_sql_command = mstrcat(sql_command, bitsandpieces, &sql_command_len, 0)) == NULL) {
	    retval = 1;
	    break;
	  }
	  else {
	    sql_command = new_sql_command;
	  }

	  token = sql_tokenize(newtoken, &sqltoken);
	  newtoken = sqltoken.next_token;
	  if (token != NULL) {
	    if ((new_buffer = mstrncpy(buffer, token, sqltoken.length, &buffer_len)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      buffer = new_buffer;
	    }
	    buffer[sqltoken.length] ='\0';

	    if ((new_sql_command = mstrcat(sql_command, buffer, &sql_command_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command = new_sql_command;
	    }
	  }
	  LOG_PRINT(LOG_DEBUG, sql_command);

	  dbires = dbi_conn_query(conn, sql_command);
	  if (!dbires) {
	    retval = 2;
	    break;
	  }

	  if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_id IN (", &sql_command1_len, 0)) == NULL) {
	    retval = 1;
	    break;
	  }
	  else {
	    sql_command1 = new_sql_command1;
	  }

	  if (dbi_result_get_numrows(dbires)) {
	    while (dbi_result_next_row(dbires) != 0) {
	      n_id = my_dbi_result_get_idval_idx(dbires, 1);
	      sprintf(the_id, ULLSPEC, (unsigned long long)n_id);
	      if ((new_sql_command1 = mstrcat(sql_command1, the_id, &sql_command1_len, 0)) == NULL) {
		retval = 1;
		dbi_result_free(dbires);
		break;
	      }
	      else {
		sql_command1 = new_sql_command1;
	      }
	      if ((new_sql_command1 = mstrcat(sql_command1, ",", &sql_command1_len, 0)) == NULL) {
		retval = 1;
		dbi_result_free(dbires);
		break;
	      }
	      else {
		sql_command1 = new_sql_command1;
	      }
	    }
	    /* remove trailing comma */
	    sql_command1[strlen(sql_command1)-1] = '\0';
	  }
	  else { /* use a dummy value that does not exist in the database */
	    if ((new_sql_command1 = mstrcat(sql_command1, "-1", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      dbi_result_free(dbires);
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }	    
	  
	  if ((new_sql_command1 = mstrcat(sql_command1, ")", &sql_command1_len, 0)) == NULL) {
	    retval = 1;
	    dbi_result_free(dbires);
	    break;
	  }
	  else {
	    sql_command1 = new_sql_command1;
	  }
	  dbi_result_free(dbires);
	}

	/*----------------------------------------------------------------*/
	/* extended notes - note citekey */
	else if (strncmp(token, ":NCK:", 5) == 0) { /* extended note citekey query */
	  strncpy(operator, &token[5], sqltoken.length-5); /* save the comparison operator */
	  operator[sqltoken.length-5] = '\0'; /* terminate string */

	  /* assemble a surrogate subselect query */
	  quoted_token = mstrdup(ptr_clrequest->username);
	  if (!quoted_token) {
	    retval = 1;
	    break;
	  }
	  if (dbi_conn_quote_string(conn, &quoted_token) == 0) {
	    retval = 1;
	    break;
	  }

	  sprintf(sql_command, "SELECT t_refdb.refdb_id FROM t_refdb, t_note, t_xnote, t_user WHERE t_note.note_user_id=t_user.user_id AND t_user.user_name=%s AND t_refdb.refdb_id=t_xnote.xref_id AND t_xnote.xnote_type=\'REFERENCE\' AND t_note.note_id=t_xnote.note_id AND t_note.note_key", quoted_token);

	  free(quoted_token);

	  /* translate operator to regexp operator */
	  if (*operator == '=') {
	    strcpy(bitsandpieces, " = ");
	  }
	  else if (*operator == '~') {
	    sprintf(bitsandpieces, " %s ", my_dbi_conn_get_cap(conn, "rlike"));
	  }
	  else if (*operator == '!' && *(operator+1) == '=') {
	    strcpy(bitsandpieces, " != ");
	  }
	  else { /* treat all other operators as non-equal */
	    sprintf(bitsandpieces, " %s ", my_dbi_conn_get_cap(conn, "not_rlike"));
	  }

	  if ((new_sql_command = mstrcat(sql_command, bitsandpieces, &sql_command_len, 0)) == NULL) {
	    retval = 1;
	    break;
	  }
	  else {
	    sql_command = new_sql_command;
	  }

	  token = sql_tokenize(newtoken, &sqltoken);
	  newtoken = sqltoken.next_token;
	  if (token != NULL) {
	    if ((new_buffer = mstrncpy(buffer, token, sqltoken.length, &buffer_len)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      buffer = new_buffer;
	    }
	    buffer[sqltoken.length] ='\0';
	    quoted_token = mstrdup(buffer);
	    if (!quoted_token) {
	      retval = 1;
	      break;
	    }
	    if (dbi_conn_quote_string(conn, &quoted_token) == 0) {
	      retval = 1;
	      break;
	    }
	    if ((new_sql_command = mstrcat(sql_command, quoted_token, &sql_command_len, 0)) == NULL) {
	      retval = 1;
	      free(quoted_token);
	      break;
	    }
	    else {
	      sql_command = new_sql_command;
	    }
	    free(quoted_token);
	  }
	  LOG_PRINT(LOG_DEBUG, sql_command);

	  dbires = dbi_conn_query(conn, sql_command);
	  if (!dbires) {
	    retval = 2;
	    break;
	  }

	  if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_id IN (", &sql_command1_len, 0)) == NULL) {
	    retval = 1;
	    break;
	  }
	  else {
	    sql_command1 = new_sql_command1;
	  }

	  if (dbi_result_get_numrows(dbires)) {
	    while (dbi_result_next_row(dbires) != 0) {
	      n_id = my_dbi_result_get_idval_idx(dbires, 1);
	      sprintf(the_id, ULLSPEC, (unsigned long long)n_id);
	      if ((new_sql_command1 = mstrcat(sql_command1, the_id, &sql_command1_len, 0)) == NULL) {
		retval = 1;
		dbi_result_free(dbires);
		break;
	      }
	      else {
		sql_command1 = new_sql_command1;
	      }
	      if ((new_sql_command1 = mstrcat(sql_command1, ",", &sql_command1_len, 0)) == NULL) {
		retval = 1;
		dbi_result_free(dbires);
		break;
	      }
	      else {
		sql_command1 = new_sql_command1;
	      }
	    }
	    /* remove trailing comma */
	    sql_command1[strlen(sql_command1)-1] = '\0';
	  }
	  else { /* use a dummy value that does not exist in the database */
	    if ((new_sql_command1 = mstrcat(sql_command1, "-1", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      dbi_result_free(dbires);
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }	    
	  
	  if ((new_sql_command1 = mstrcat(sql_command1, ")", &sql_command1_len, 0)) == NULL) {
	    retval = 1;
	    dbi_result_free(dbires);
	    break;
	  }
	  else {
	    sql_command1 = new_sql_command1;
	  }
	  dbi_result_free(dbires);
	}

	/*----------------------------------------------------------------*/
	/* Personal info: reprint, notes, availability */
	else if (strncmp(token, ":RP:", 4) == 0 || /* reprint query */
		 strncmp(token, ":N1:", 4) == 0 || /* notes query */
		 strncmp(token, ":AV:", 4) == 0) { /* availability query */
	  strncpy(operator, &token[4], sqltoken.length-4); /* save the comparison operator */
	  operator[sqltoken.length-4] = '\0'; /* terminate string */

	  /* assemble a surrogate subselect query */
	  if (*operator == '=') {
	    if (strncmp(token, ":RP:", 4) == 0) {
	      strcpy(sql_command, "SELECT t_xuser.refdb_id FROM t_refdb, t_xuser WHERE t_refdb.refdb_id=t_xuser.refdb_id AND t_xuser.xuser_reprint = ");
	    }
	    else if (strncmp(token, ":N1:", 4) == 0) {
	      strcpy(sql_command, "SELECT t_xuser.refdb_id FROM t_refdb, t_xuser WHERE t_refdb.refdb_id=t_xuser.refdb_id AND t_xuser.xuser_notes = ");
	    }
	    else {
	      strcpy(sql_command, "SELECT t_xuser.refdb_id FROM t_refdb, t_xuser WHERE t_refdb.refdb_id=t_xuser.refdb_id AND t_xuser.xuser_avail = ");
	    }
	  }
	  else if (*operator == '~') {
	    if (strncmp(token, ":RP:", 4) == 0) {
	      sprintf(sql_command, "SELECT t_xuser.refdb_id FROM t_refdb, t_xuser WHERE t_refdb.refdb_id=t_xuser.refdb_id AND t_xuser.xuser_reprint %s ", my_dbi_conn_get_cap(conn, "rlike"));
	    }
	    else if (strncmp(token, ":N1:", 4) == 0) {
	      sprintf(sql_command, "SELECT t_xuser.refdb_id FROM t_refdb, t_xuser WHERE t_refdb.refdb_id=t_xuser.refdb_id AND t_xuser.xuser_notes %s ", my_dbi_conn_get_cap(conn, "rlike"));
	    }
	    else {
	      sprintf(sql_command, "SELECT t_xuser.refdb_id FROM t_refdb, t_xuser WHERE t_refdb.refdb_id=t_xuser.refdb_id AND t_xuser.xuser_avail %s ", my_dbi_conn_get_cap(conn, "rlike"));
	    }
	  }
	  else if (*operator == '!' && *(operator+1) == '=') {
	    if (strncmp(token, ":RP:", 4) == 0) {
	      strcpy(sql_command, "SELECT t_xuser.refdb_id FROM t_refdb, t_xuser WHERE t_refdb.refdb_id=t_xuser.refdb_id AND t_xuser.xuser_reprint != ");
	    }
	    else if (strncmp(token, ":N1:", 4) == 0) {
	      strcpy(sql_command, "SELECT t_xuser.refdb_id FROM t_refdb, t_xuser WHERE t_refdb.refdb_id=t_xuser.refdb_id AND t_xuser.xuser_notes != ");
	    }
	    else {
	      strcpy(sql_command, "SELECT t_xuser.refdb_id FROM t_refdb, t_xuser WHERE t_refdb.refdb_id=t_xuser.refdb_id AND t_xuser.xuser_avail != ");
	    }
	  }
	  else {
	    if (strncmp(token, ":RP:", 4) == 0) {
	      sprintf(sql_command, "SELECT t_xuser.refdb_id FROM t_refdb, t_xuser WHERE t_refdb.refdb_id=t_xuser.refdb_id AND t_xuser.xuser_reprint %s ", my_dbi_conn_get_cap(conn, "not_rlike"));
	    }
	    else if (strncmp(token, ":N1:", 4) == 0) {
	      sprintf(sql_command, "SELECT t_xuser.refdb_id FROM t_refdb, t_xuser WHERE t_refdb.refdb_id=t_xuser.refdb_id AND t_xuser.xuser_notes %s ", my_dbi_conn_get_cap(conn, "not_rlike"));
	    }
	    else {
	      sprintf(sql_command, "SELECT t_xuser.refdb_id FROM t_refdb, t_xuser WHERE t_refdb.refdb_id=t_xuser.refdb_id AND t_xuser.xuser_avail %s ", my_dbi_conn_get_cap(conn, "not_rlike"));
	    }
	  }

	  token = sql_tokenize(newtoken, &sqltoken);
	  newtoken = sqltoken.next_token;
	  if (token != NULL) {
	    if ((new_buffer = mstrncpy(buffer, token, sqltoken.length, &buffer_len)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      buffer = new_buffer;
	    }
	    buffer[sqltoken.length] ='\0';
	    quoted_token = mstrdup(buffer);
	    if (!quoted_token) {
	      retval = 1;
	      break;
	    }
	    if (dbi_conn_quote_string(conn, &quoted_token) == 0) {
	      retval = 1;
	      break;
	    }
	    if ((new_sql_command = mstrcat(sql_command, quoted_token, &sql_command_len, 0)) == NULL) {
	      retval = 1;
	      free(quoted_token);
	      break;
	    }
	    else {
	      sql_command = new_sql_command;
	    }
	    free(quoted_token);
	  }
	  LOG_PRINT(LOG_DEBUG, sql_command);

	  /* dbires will contain a list of all refdb_ids that match
	     the above publication subquery */
	  dbires = dbi_conn_query(conn, sql_command);
	  if (!dbires) {
	    retval = 2;
	    break;
	  }

	  if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_id IN (", &sql_command1_len, 0)) == NULL) {
	    retval = 1;
	    break;
	  }
	  else {
	    sql_command1 = new_sql_command1;
	  }

	  n_havesome = 0;
	  while (dbi_result_next_row(dbires) != 0) {
	    n_havesome = 1;
	    n_id = my_dbi_result_get_idval_idx(dbires, 1);
	    sprintf(the_id, ULLSPEC, (unsigned long long)n_id);
	    if ((new_sql_command1 = mstrcat(sql_command1, the_id, &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      dbi_result_free(dbires);
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	    if ((new_sql_command1 = mstrcat(sql_command1, ",", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      dbi_result_free(dbires);
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  if (!n_havesome) { /* use a dummy value that does not exist in the database */
	    if ((new_sql_command1 = mstrcat(sql_command1, "-1,", &sql_command1_len, 0)) == NULL) { /* trailing comma will be overwritten below */
	      retval = 1;
	      dbi_result_free(dbires);
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }	    

	  dbi_result_free(dbires);
	  if (retval) {
	    break;
	  }
	  if ((new_sql_command1 = mstrcat(sql_command1, ")", &sql_command1_len, 1)) == NULL) { /* overwrites trailing comma */
	    retval = 1;
	    break;
	  }
	  else {
	    sql_command1 = new_sql_command1;
	  }
	}

	/*----------------------------------------------------------------*/
	/* simple queries */
	else { /* all simple t_refdb queries */
	  strncpy(operator, &token[4], sqltoken.length-4); /* save the comparison operator */
	  operator[sqltoken.length-4] = '\0'; /* terminate string */
	  
	  /* initialize alphanumeric indicator */
	  n_isalpha = 1;

	  if (strncmp(token, ":TY:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_type", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":PY:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_pubyear", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	      n_isalpha = 0;
	    }
	  }
	  else if (strncmp(token, ":Y2:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_secyear", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	      n_isalpha = 0;
	    }
	  }
	  else if (strncmp(token, ":ID:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_id", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	      n_isalpha = 0;
	    }
	  }
	  else if (strncmp(token, ":CK:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_citekey", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":TI:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_title", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":T2:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_booktitle", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":T3:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_title_series", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":SN:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_issn", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":N2:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_abstract", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":SP:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_startpage", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":EP:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_endpage", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":VL:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_volume", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":IS:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_chapnumber", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }

	  /* todo: must retrieve data from separate tables */
	  else if (strncmp(token, ":UR:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_url", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":AD:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_address", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":CY:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_city", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":PB:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_publisher", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":U1:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_user1", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":U2:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_user2", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":U3:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_user3", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":U4:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_user4", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":U5:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_user5", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":M1:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_misc1", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":M2:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_misc2", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":M3:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_misc3", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":L1:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_linkpdf", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":L2:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_linkfull", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":L3:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_linkrel", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else if (strncmp(token, ":L4:", 4) == 0) {
	    if ((new_sql_command1 = mstrcat(sql_command1, "t_refdb.refdb_linkimg", &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }

	  /* we have to treat two types of columns: numerical columns
	     can reuse the operator directly. Alphanumeric columns need
	     some intelligent translation into a regexp-based query */
	  if (n_isalpha) {
	    if (*operator == '=') {
 	      strcpy(bitsandpieces, " = ");
	    }
	    else if (*operator == '~') {
	      sprintf(bitsandpieces, " %s ", my_dbi_conn_get_cap(conn, "rlike"));
	    }
	    else if (*operator == '!' && *(operator+1) == '=') {
 	      sprintf(bitsandpieces, " != ");
	    }
	    else { /* treat all other operators as non-equal */
	      sprintf(bitsandpieces, " %s ", my_dbi_conn_get_cap(conn, "not_rlike"));
	    }

	    if ((new_sql_command1 = mstrcat(sql_command1, bitsandpieces, &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  else {
	    if ((new_sql_command1 = mstrcat(sql_command1, operator, &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	  }
	  token = sql_tokenize(newtoken, &sqltoken);
	  newtoken = sqltoken.next_token;
	  if (token != NULL) {
/*  	    printf("%s:%d\n", token, retval); */
/*  	    printf("%s\n", newtoken); */
	    if ((new_buffer = mstrncpy(buffer, token, sqltoken.length, &buffer_len)) == NULL) {
	      retval = 1;
	      break;
	    }
	    else {
	      buffer = new_buffer;
	    }
	    buffer[sqltoken.length] ='\0';
	    quoted_token = mstrdup(buffer);
	    if (!quoted_token) {
	      retval = 1;
	      break;
	    }
	    if (dbi_conn_quote_string(conn, &quoted_token) == 0) {
	      retval = 1;
	      break;
	    }
	    if ((new_sql_command1 = mstrcat(sql_command1, quoted_token, &sql_command1_len, 0)) == NULL) {
	      retval = 1;
	      free(quoted_token);
	      break;
	    }
	    else {
	      sql_command1 = new_sql_command1;
	    }
	    free(quoted_token);
	  }
	}
      } /* if (sqltoken.type != 4) / else */
    } /* if (token != NULL) */
  } /* while (newtoken != NULL) */

  if (retval) {
    if (retval == 1) {
      send_status(ptr_clrequest->fd, 801, TERM_NO);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
    }
    else if (retval == 2) {
      send_status(ptr_clrequest->fd, 234, TERM_NO);
      LOG_PRINT(LOG_WARNING, get_status_msg(234));
    }
/*      printf("retval detected\n"); */
    dbi_conn_close(conn);
    delete_all_lilimem(&sentinel);
    if (conv_descriptor) {
      iconv_close(conv_descriptor);
    }
    return 1;
  }

  /* now assemble the SQL query string proper */
/*    printf("assembling query\n"); */
  /* 0  refdb_id             ID  - (if numeric)
     1  refdb_type           TY  - 
     2  refdb_pubyear        PY  - (partial)
     3  refdb_startpage      SP  - 
     4  refdb_endpage        EP  - 
     5  refdb_abstract       N2  - 
     6  refdb_title          TI  - 
     7  refdb_volume         VL  - 
     8  refdb_issue          CP  - 
     9  refdb_booktitle      BT  - 
    10  refdb_city           CY  - 
    11  refdb_publisher      PB  - 
    12  refdb_title_series   T3  - 
    13  refdb_address        AD  - 
    deleted14  refdb_url            UR  - 
    14  refdb_issn           SN  - 
    15  refdb_periodical_id  JO  - (indirect)
    16  refdb_pyother_info   PY  - (partial)
    17  refdb_secyear        Y2  - (partial)
    18  refdb_secother_info  Y2  - (partial)
    19  refdb_user1          U1  - 
    20  refdb_user2          U2  - 
    21  refdb_user3          U3  - 
    22  refdb_user4          U4  - 
    23  refdb_user5          U5  - 
    24  refdb_misc1          M1  - 
    25  refdb_misc2          M2  - 
    26  refdb_misc3          M3  - 
    27  refdb_citekey        ID  - (if non-numeric)
    deleted 29  refdb_linkpdf        L1  - 
    deleted 30  refdb_linkfull       L2  - 
    deleted 31  refdb_linkrel        L3  - 
    deleted 32  refdb_linkimg        L4  - 
*/
  strcpy(sql_command, "SELECT DISTINCT t_refdb.refdb_id, t_refdb.refdb_type, t_refdb.refdb_pubyear, t_refdb.refdb_startpage, t_refdb.refdb_endpage, t_refdb.refdb_abstract, t_refdb.refdb_title, t_refdb.refdb_volume, t_refdb.refdb_issue, t_refdb.refdb_booktitle, t_refdb.refdb_city, t_refdb.refdb_publisher, t_refdb.refdb_title_series, t_refdb.refdb_address, t_refdb.refdb_issn, t_refdb.refdb_periodical_id, t_refdb.refdb_pyother_info, t_refdb.refdb_secyear, t_refdb.refdb_secother_info, t_refdb.refdb_user1, t_refdb.refdb_user2, t_refdb.refdb_user3, t_refdb.refdb_user4, t_refdb.refdb_user5, t_refdb.refdb_misc1, t_refdb.refdb_misc2, t_refdb.refdb_misc3, t_refdb.refdb_citekey FROM t_refdb WHERE ");


  if (ref_format != 1 || !strstr(ptr_biblio_info->format_string, "NOHOLES")) {
    if ((new_sql_command = mstrcat(sql_command, "refdb_type!='DUMMY' AND ", &sql_command_len, 0)) == NULL) {
      send_status(ptr_clrequest->fd, 801, TERM_NO);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      dbi_conn_close(conn);
      delete_all_lilimem(&sentinel);
      if (conv_descriptor) {
	iconv_close(conv_descriptor);
      }
      return 1;
    }
    else {
      sql_command = new_sql_command;
    }
  }

  if ((new_sql_command = mstrcat(sql_command, sql_command1, &sql_command_len, 0)) == NULL) {
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    dbi_conn_close(conn);
    delete_all_lilimem(&sentinel);
    if (conv_descriptor) {
      iconv_close(conv_descriptor);
    }
    return 1;
  }
  else {
    sql_command = new_sql_command;
  }

  /* see whether we need the user's list instead of the whole list */
  if (n_privatelist) {
    /* get reference list for current user */
    strcpy(sql_command1, "SELECT t_xuser.refdb_id FROM t_user, t_xuser WHERE t_user.user_id=t_xuser.user_id AND t_user.user_name=");
    quoted_token = mstrdup(ptr_clrequest->username);
    if (!quoted_token) {
      send_status(ptr_clrequest->fd, 801, TERM_NO);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      dbi_conn_close(conn);
      delete_all_lilimem(&sentinel);
      if (conv_descriptor) {
	iconv_close(conv_descriptor);
      }
      return 1;
    }
    if (dbi_conn_quote_string(conn, &quoted_token) == 0) {
      dbi_conn_close(conn);
      delete_all_lilimem(&sentinel);
    }
    strcat(sql_command1, quoted_token);
    free(quoted_token);

    LOG_PRINT(LOG_DEBUG, sql_command1);

    dbires = dbi_conn_query(conn, sql_command1);
    if (!dbires) {
      send_status(ptr_clrequest->fd, 234, TERM_NO);
      LOG_PRINT(LOG_WARNING, get_status_msg(234));
      dbi_conn_close(conn);
      delete_all_lilimem(&sentinel);
      if (conv_descriptor) {
	iconv_close(conv_descriptor);
      }
      return 1;
    }
    
    if ((new_sql_command = mstrcat(sql_command, " AND t_refdb.refdb_id IN (", &sql_command_len, 0)) == NULL) {
      send_status(ptr_clrequest->fd, 801, TERM_NO);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      dbi_result_free(dbires);
      dbi_conn_close(conn);
      delete_all_lilimem(&sentinel);
      if (conv_descriptor) {
	iconv_close(conv_descriptor);
      }
      return 1;
    }
    else {
      sql_command = new_sql_command;
    }

    if (dbi_result_get_numrows(dbires)) {
      while (dbi_result_next_row(dbires) != 0) {
	n_id = my_dbi_result_get_idval_idx(dbires, 1);
	sprintf(the_id, ULLSPEC, (unsigned long long)n_id);
	if ((new_sql_command = mstrcat(sql_command, the_id, &sql_command_len, 0)) == NULL) {
	  retval = 1;
	  send_status(ptr_clrequest->fd, 801, TERM_NO);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  dbi_result_free(dbires);
	  dbi_conn_close(conn);
	  delete_all_lilimem(&sentinel);
	  if (conv_descriptor) {
	    iconv_close(conv_descriptor);
	  }
	  return 1;
	}
	else {
	  sql_command = new_sql_command;
	}
	if ((new_sql_command = mstrcat(sql_command, ",", &sql_command_len, 0)) == NULL) {
	  retval = 1;
	  send_status(ptr_clrequest->fd, 801, TERM_NO);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  dbi_result_free(dbires);
	  dbi_conn_close(conn);
	  delete_all_lilimem(&sentinel);
	  if (conv_descriptor) {
	    iconv_close(conv_descriptor);
	  }
	  return 1;
	}
	else {
	  sql_command = new_sql_command;
	}
      } /* end while */
    }
    else { /* if no rows are available */
      if ((new_sql_command = mstrcat(sql_command, "0,", &sql_command_len, 0)) == NULL) {
	retval = 1;
	send_status(ptr_clrequest->fd, 801, TERM_NO);
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	dbi_result_free(dbires);
	dbi_conn_close(conn);
	delete_all_lilimem(&sentinel);
	if (conv_descriptor) {
	  iconv_close(conv_descriptor);
	}
	return 1;
      }
      else {
	sql_command = new_sql_command;
      }
    } /* end if have rows */

    dbi_result_free(dbires);
    
    /* overwrite the trailing comma */
    if (retval || ((new_sql_command = mstrcat(sql_command, ")", &sql_command_len, 1)) == NULL)) {
      send_status(ptr_clrequest->fd, 801, TERM_NO);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      dbi_conn_close(conn);
      delete_all_lilimem(&sentinel);
      if (conv_descriptor) {
	iconv_close(conv_descriptor);
      }
      return 1;
    }
    else {
      sql_command = new_sql_command;
    }
  }

  /* sort the output */
  if (strncmp(ptr_biblio_info->sort_string, "PY", 2) == 0) {
    if (retval || ((new_sql_command = mstrcat(sql_command, " ORDER BY t_refdb.refdb_pubyear", &sql_command_len, 0)) == NULL)) {
      send_status(ptr_clrequest->fd, 801, TERM_NO);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      dbi_conn_close(conn);
      delete_all_lilimem(&sentinel);
      if (conv_descriptor) {
	iconv_close(conv_descriptor);
      }
      return 1;
    }
    else {
      sql_command = new_sql_command;
    }
  }
  else {
    if (retval || ((new_sql_command = mstrcat(sql_command, " ORDER BY t_refdb.refdb_id", &sql_command_len, 0)) == NULL)) {
      send_status(ptr_clrequest->fd, 801, TERM_NO);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      dbi_conn_close(conn);
      delete_all_lilimem(&sentinel);
      if (conv_descriptor) {
	iconv_close(conv_descriptor);
      }
      return 1;
    }
    else {
      sql_command = new_sql_command;
    }
  }

  LOG_PRINT(LOG_DEBUG, sql_command);

  rendinfo.dbires = dbi_conn_query(conn, sql_command);
  if (!rendinfo.dbires) {
    send_status(ptr_clrequest->fd, 234, TERM_NO);
    LOG_PRINT(LOG_WARNING, get_status_msg(234));
    dbi_conn_close(conn);
    delete_all_lilimem(&sentinel);
    if (conv_descriptor) {
      iconv_close(conv_descriptor);
    }
    return 1;
  }

  /* create a "header" if necessary */
  /* prepare_render resets sql_command */

  /* fill in variable elements of structure */
  rendinfo.ptr_ref = &sql_command;
  rendinfo.ptr_ref_len = &sql_command_len;
  if (!*(rendinfo.ptr_biblio_info->encoding)) {
    rendinfo.ptr_biblio_info->encoding = dbi_conn_get_encoding(conn);
  }

  /* real output starts here */
  if (prepare_render(&rendinfo)) {
    send_status(ptr_clrequest->fd, 801, TERM_NO);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    dbi_result_free(rendinfo.dbires);
    dbi_conn_close(conn);
    delete_all_lilimem(&sentinel);
    if (conv_descriptor) {
      iconv_close(conv_descriptor);
    }
    return 1;
  }

  if (dbi_result_get_numrows(rendinfo.dbires)) {
    /* fetch all articles we're interested in */
    while (dbi_result_next_row(rendinfo.dbires) != 0) {
      nref_counter++;
      /* fill in variable elements of structure */
      rendinfo.ptr_ref = &sql_command;
      rendinfo.ptr_ref_len = &sql_command_len;
      rendinfo.nref_counter = nref_counter;
      
      /* todo: count query errors and such as failed datasets, bail out only
	 after irrecoverable errors */
      if ((render_res = render_ref(&rendinfo)) != 0) {
	if (render_res == 801) {
	  retval = 1;
	  send_status(ptr_clrequest->fd, render_res, TERM_NO);
	  dbi_result_free(rendinfo.dbires);
	  dbi_conn_close(conn);
	  delete_all_lilimem(&sentinel);
	  if (conv_descriptor) {
	    iconv_close(conv_descriptor);
	  }
	  return 0;
	}
	else {
	  ptr_addresult->failure++;
	}
      }

      /* run a character encoding conversion if required */
      if (conv_descriptor && *sql_command) {
	size_t inlength;
	size_t outlength;
	char* my_sql_command = NULL; /* this ptr will be modified by iconv() */
	char* my_sql_command_start = NULL; /* records initial state of my_elvalue */
	const char* my_instring = NULL; /* this ptr will be modified by iconv() */
	/* strlen should be ok here as this can't be a multibyte encoding */
	inlength = strlen(sql_command)/*  + 1 */;
	/* with the encodings supported by our database engines, the converted
	   string can't be longer than six times the input string */
	/* todo: is this assumption correct? */
	outlength = 6*inlength;
	
	if ((my_sql_command = (char*)calloc(outlength, sizeof(char))) == NULL) {
	  retval = 1;
	  send_status(ptr_clrequest->fd, 801, TERM_NO);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  dbi_result_free(rendinfo.dbires);
	  dbi_conn_close(conn);
	  delete_all_lilimem(&sentinel);
	  if (conv_descriptor) {
	    iconv_close(conv_descriptor);
	  }
	  return 0;
	}

	/* keep start of the converted string */
	my_sql_command_start = my_sql_command;

	/* variable will be modified by iconv, so don't use original */
	my_instring = (const char*)sql_command;

	/* now actually do the conversion */
	if (iconv(conv_descriptor, &my_instring, &inlength, &my_sql_command, &outlength) == (size_t)(-1)) {
	  if (errno == EILSEQ) {
	    LOG_PRINT(LOG_WARNING, "iconv: invalid input character sequence");
	  }
	  else if (errno == E2BIG) {
	    LOG_PRINT(LOG_WARNING, "iconv: output buffer too small");
	  }
	  else if (errno == EINVAL) {
	    LOG_PRINT(LOG_WARNING, "iconv: incomplete input character");
	  }
	  
	  retval = 1;
	  send_status(ptr_clrequest->fd, 702, TERM_NO);
	  dbi_result_free(rendinfo.dbires);
	  dbi_conn_close(conn);
	  delete_all_lilimem(&sentinel);
	  if (conv_descriptor) {
	    iconv_close(conv_descriptor);
	  }
	  return 0;
	}
	/* else: conversion went ok. We free the original string and replace
	   it with the converted copy */
	if (sql_command) {
	  free(sql_command);
	}
	sql_command = my_sql_command_start;
	sql_command_len = outlength;
	result_len = (size_t)(my_sql_command - my_sql_command_start - 1);
      }
      else { /* no conversion required */
	result_len = strlen(sql_command);
      }

      /* send ok status, then the terminated result string */
      send_status(ptr_clrequest->fd, 404, TERM_NO);
      iwrite(ptr_clrequest->fd, sql_command, result_len); 
      iwrite(ptr_clrequest->fd, cs_term, TERM_LEN);
      ptr_addresult->success++;
      
      /* reset buffer string */
      sql_command[0] = '\0';
      
      /* read client response */
      cs_status = read_status(ptr_clrequest->fd);

      if (cs_status) {
	break;
      }
    } /* end while */
  }
  else {
    /* no datasets, return an empty body */
      send_status(ptr_clrequest->fd, 402, TERM_NO);
      iwrite(ptr_clrequest->fd, cs_term, TERM_LEN);
  }

  if (conv_descriptor) {
    iconv_close(conv_descriptor);
  }

  if (ptr_addresult->success) {
    /* create a "footer" if necessary */
    /* fill in variable elements of structure */
/*     rendinfo.ptr_ref = &sql_command; */
/*     rendinfo.ptr_ref_len = &sql_command_len; */

    if ((render_res = finish_render(&rendinfo)) != 0) {
      retval = 1;
      send_status(ptr_clrequest->fd, 801, TERM_NO);
      dbi_result_free(rendinfo.dbires);
      dbi_conn_close(conn);
      delete_all_lilimem(&sentinel);
      return 0;
    }
    send_status(ptr_clrequest->fd, 402, TERM_NO);
    tiwrite(ptr_clrequest->fd, sql_command, TERM_YES); 
  }

  dbi_result_free(rendinfo.dbires);
  dbi_conn_close(conn);
  delete_all_lilimem(&sentinel);

  return retval;
}

